import pytest

from brian2 import *
from brian2.devices.device import reinit_and_delete
from brian2.tests.utils import assert_allclose
from brian2.utils.caching import _hashable


@pytest.mark.codegen_independent
def test_timedarray_direct_use():
    ta = TimedArray(np.linspace(0, 10, 11), 1 * ms)
    assert ta(-1 * ms) == 0
    assert ta(5 * ms) == 5
    assert ta(10 * ms) == 10
    assert ta(15 * ms) == 10
    ta = TimedArray(np.linspace(0, 10, 11) * amp, 1 * ms)
    assert ta(-1 * ms) == 0 * amp
    assert ta(5 * ms) == 5 * amp
    assert ta(10 * ms) == 10 * amp
    assert ta(15 * ms) == 10 * amp
    ta2d = TimedArray((np.linspace(0, 11, 12) * amp).reshape(4, 3), 1 * ms)
    assert ta2d(-1 * ms, 0) == 0 * amp
    assert ta2d(0 * ms, 0) == 0 * amp
    assert ta2d(0 * ms, 1) == 1 * amp
    assert ta2d(1 * ms, 1) == 4 * amp
    assert_allclose(ta2d(1 * ms, [0, 1, 2]), [3, 4, 5] * amp)
    assert_allclose(ta2d(15 * ms, [0, 1, 2]), [9, 10, 11] * amp)


@pytest.mark.standalone_compatible
def test_timedarray_semantics():
    # Make sure that timed arrays are interpreted as specifying the values
    # between t and t+dt (not between t-dt/2 and t+dt/2 as in Brian1)
    ta = TimedArray(array([0, 1]), dt=0.4 * ms)
    G = NeuronGroup(1, "value = ta(t) : 1", dt=0.1 * ms)
    mon = StateMonitor(G, "value", record=0)
    run(0.8 * ms)
    assert_allclose(mon[0].value, [0, 0, 0, 0, 1, 1, 1, 1])
    assert_allclose(mon[0].value, ta(mon.t))


@pytest.mark.standalone_compatible
def test_timedarray_no_units():
    ta = TimedArray(np.arange(10), dt=0.1 * ms)
    G = NeuronGroup(1, "value = ta(t) + 1: 1", dt=0.1 * ms)
    mon = StateMonitor(G, "value", record=True, dt=0.1 * ms)
    run(1.1 * ms)
    assert_allclose(mon[0].value_, np.clip(np.arange(len(mon[0].t)), 0, 9) + 1)


@pytest.mark.standalone_compatible
def test_timedarray_with_units():
    ta = TimedArray(np.arange(10) * amp, dt=0.1 * ms)
    G = NeuronGroup(1, "value = ta(t) + 2*nA: amp", dt=0.1 * ms)
    mon = StateMonitor(G, "value", record=True, dt=0.1 * ms)
    run(1.1 * ms)
    assert_allclose(
        mon[0].value, np.clip(np.arange(len(mon[0].t)), 0, 9) * amp + 2 * nA
    )


@pytest.mark.standalone_compatible
def test_timedarray_2d():
    # 4 time steps, 3 neurons
    ta2d = TimedArray(np.arange(12).reshape(4, 3), dt=0.1 * ms)
    G = NeuronGroup(3, "value = ta2d(t, i) + 1: 1", dt=0.1 * ms)
    mon = StateMonitor(G, "value", record=True, dt=0.1 * ms)
    run(0.5 * ms)
    assert_allclose(mon[0].value_, np.array([0, 3, 6, 9, 9]) + 1)
    assert_allclose(mon[1].value_, np.array([1, 4, 7, 10, 10]) + 1)
    assert_allclose(mon[2].value_, np.array([2, 5, 8, 11, 11]) + 1)


@pytest.mark.codegen_independent
def test_timedarray_incorrect_use():
    ta = TimedArray(np.linspace(0, 10, 11), 1 * ms)
    ta2d = TimedArray((np.linspace(0, 11, 12) * amp).reshape(4, 3), 1 * ms)
    G = NeuronGroup(3, "I : amp")
    with pytest.raises(ValueError):
        setattr(G, "I", "ta2d(t)*amp")
    with pytest.raises(ValueError):
        setattr(G, "I", "ta(t, i)*amp")
    with pytest.raises(ValueError):
        setattr(G, "I", "ta()*amp")
    with pytest.raises(ValueError):
        setattr(G, "I", "ta*amp")


@pytest.mark.standalone_compatible
def test_timedarray_no_upsampling():
    # Test a TimedArray where no upsampling is necessary because the monitor's
    # dt is bigger than the TimedArray's
    ta = TimedArray(np.arange(10), dt=0.01 * ms)
    G = NeuronGroup(1, "value = ta(t): 1", dt=0.1 * ms)
    mon = StateMonitor(G, "value", record=True, dt=1 * ms)
    run(2.1 * ms)
    assert_allclose(mon[0].value, [0, 9, 9])


# @pytest.mark.standalone_compatible  # see FIXME comment below
def test_long_timedarray():
    """
    Use a very long timedarray (with a big dt), where the upsampling can lead
    to integer overflow.
    """
    ta = TimedArray(np.arange(16385), dt=1 * second)
    G = NeuronGroup(1, "value = ta(t) : 1")
    mon = StateMonitor(G, "value", record=True)
    net = Network(G, mon)
    # We'll start the simulation close to the critical boundary
    # FIXME: setting the time like this does not work for standalone
    net.t_ = float(16384 * second - 5 * ms)
    net.run(10 * ms)
    assert_allclose(mon[0].value[mon.t < 16384 * second], 16383)
    assert_allclose(mon[0].value[mon.t >= 16384 * second], 16384)


def test_timedarray_repeated_use():
    # Check that recreating a TimedArray with different values does work
    # correctly (no issues with caching)
    values = np.array([[1, 2, 3], [2, 4, 6]])
    for run_idx in range(2):
        ta = TimedArray(values[run_idx], dt=defaultclock.dt, name="ta")
        G = NeuronGroup(1, "dx/dt = ta(t)/dt : 1", name="G")
        run(3 * defaultclock.dt)
        assert G.x[0] == 6 * (run_idx + 1)


if __name__ == "__main__":
    test_timedarray_direct_use()
    test_timedarray_semantics()
    test_timedarray_no_units()
    test_timedarray_with_units()
    test_timedarray_2d()
    test_timedarray_incorrect_use()
    test_timedarray_no_upsampling()
    test_long_timedarray()
    test_timedarray_repeated_use()



"""
Update CITATION.cff and README.md with the latest Zenodo and Software Heritage records
"""
import os
import re
import subprocess

import pyaml
import requests
import yaml

basedir = os.path.abspath(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
)

# Get the latest tag from git
tag = (
    subprocess.check_output(["git", "describe", "--tags", "--abbrev=0"])
    .strip()
    .decode("utf-8")
)

print("Latest tag/release:", tag)
ZENODO_API = "https://zenodo.org/api"
# DOI always linking to the latest version (escaped slash for elasticsearch syntax)
CONCEPT_DOI = r"10.5281\/zenodo.654861"

print("Searching for Zenodo record...")
# Get the latest version via concept doi
# The Zenodo search guide says that one can directly search by "version:...", but it does not work (BAD REQUEST)
r = requests.get(ZENODO_API + "/records", params={"q": f"conceptdoi:{CONCEPT_DOI}"})
if not r.ok:
    raise RuntimeError("Request failed: ", r.reason)
data = r.json()
assert data["hits"]["total"] == 1
latest_zenodo_id = data["hits"]["hits"][0]["id"]

# Get all versions
r = requests.get(ZENODO_API + f"/records/{latest_zenodo_id}/versions")
if not r.ok:
    raise RuntimeError("Request failed: ", r.reason)
data = r.json()
versions = data["hits"]["hits"]
latest_version = [v for v in versions if v["metadata"]["version"] == tag]
if latest_version:
    zenodo_record = latest_version[0]
    print("Found Zenodo record for version", tag)
else:
    raise RuntimeError("No Zenodo record found for version " + tag)

print("Searching for SWH record...")

# Find Software Heritage
resp = requests.get(
    "https://archive.softwareheritage.org/api/1/origin/https://github.com/brian-team/brian2/get/"
)
if not resp.ok:
    raise RuntimeError("Request failed: ", resp.reason)
data = resp.json()
visits_url = data["origin_visits_url"]
resp = requests.get(visits_url)
if not resp.ok:
    raise RuntimeError("Request failed: ", resp.reason)
data = resp.json()
latest_visit = sorted(data, key=lambda x: x["date"], reverse=True)[0]
snapshot_url = latest_visit["snapshot_url"]
resp = requests.get(snapshot_url)
if not resp.ok:
    raise RuntimeError("Request failed: ", resp.reason)
data = resp.json()
swh_record = data["branches"].get(f"refs/tags/{tag}")

if swh_record:
    print("Found SWH record for version", tag)
else:
    swh_record = None
    print("No SWH record found for version", tag)

print("Updating CITATION.cff and README.md...")
with open(os.path.join(basedir, "CITATION.cff"), "r") as f:
    citation_cff = yaml.load(f, Loader=yaml.SafeLoader)
for identifier in citation_cff["identifiers"]:
    if zenodo_record and identifier["description"].startswith(
        "This is the archived snapshot of version"
    ):
        identifier["value"] = zenodo_record["metadata"]["doi"]
        identifier["description"] = (
            f"This is the archived snapshot of version {tag} of Brian 2"
        )
    if swh_record and identifier["type"] == "swh":
        identifier["value"] = f"swh:1:rel:{swh_record['target']}"
        identifier["description"] = (
            f"Software Heritage identifier for version {tag} of Brian 2"
        )

with open(os.path.join(basedir, "CITATION.cff"), "w") as f:
    pyaml.dump(citation_cff, f, sort_keys=False)

with open(os.path.join(basedir, "README.md"), "r") as f:
    readme = f.read()

if zenodo_record:
    # Replace the old DOI with the new one
    readme = re.sub(
        r"\[!\[DOI\]\(https://zenodo.org/badge/DOI/.*\.svg\)\]\(https://zenodo.org/doi/.*\)",
        f"[![DOI](https://zenodo.org/badge/DOI/{zenodo_record['metadata']['doi']}.svg)](https://zenodo.org/doi/{zenodo_record['metadata']['doi']})",
        readme,
    )

if swh_record:
    # Replace the old SWH badge with the new one
    # [![Software Heritage (release)](https://archive.softwareheritage.org/badge/swh:1:rel:2d4c5c8c8a6d2318332889df93ab74aef53e2c61/)](https://archive.softwareheritage.org/swh:1:rel:2d4c5c8c8a6d2318332889df93ab74aef53e2c61;origin=https://github.com/brian-team/brian2;visit=swh:1:snp:a90ab7416901a9c5cf6f56d68b3455c65d322afc)
    readme = re.sub(
        r"\[!\[Software Heritage \(release\)\]\(https://archive.softwareheritage.org/badge/swh:1:rel:.*\)\]\(https://archive.softwareheritage.org/swh:1:rel:.*;origin=.*\)",
        f"[![Software Heritage (release)](https://archive.softwareheritage.org/badge/swh:1:rel:{swh_record['target']}/)](https://archive.softwareheritage.org/swh:1:rel:{swh_record['target']};origin=https://github.com/brian-team/brian2;visit=swh:1:snp:{latest_visit['snapshot']})",
        readme,
    )

with open(os.path.join(basedir, "README.md"), "w") as f:
    f.write(readme)



{# USES_VARIABLES { _group_idx } #}
{# ALLOWS_SCALAR_WRITE #}
{% extends 'common_group.pyx' %}

{% block maincode %}

    cdef size_t _target_idx
    
    _vectorisation_idx = 1
    
    {{scalar_code|autoindent}}
    
    for _idx_group_idx in range(_num{{_group_idx}}):
        _idx = {{_group_idx}}[_idx_group_idx]
        _vectorisation_idx = _idx
        _target_idx = _idx
        
        {{vector_code|autoindent}}

{% endblock %}



import platform

import numpy

from brian2.core.base import BrianObjectException
from brian2.core.functions import Function
from brian2.core.preferences import BrianPreference, prefs
from brian2.core.variables import (
    ArrayVariable,
    AuxiliaryVariable,
    DynamicArrayVariable,
    Subexpression,
)
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import get_identifiers

from ...codeobject import check_compiler_kwds, constant_or_scalar
from ...cpp_prefs import get_compiler_and_args
from ...generators.cython_generator import (
    CythonCodeGenerator,
    get_cpp_dtype,
    get_numpy_dtype,
)
from ...targets import codegen_targets
from ...templates import Templater
from ..numpy_rt import NumpyCodeObject
from .extension_manager import cython_extension_manager

__all__ = ["CythonCodeObject"]


logger = get_logger(__name__)

# Preferences
prefs.register_preferences(
    "codegen.runtime.cython",
    "Cython runtime codegen preferences",
    multiprocess_safe=BrianPreference(
        default=True,
        docs="""
        Whether to use a lock file to prevent simultaneous write access
        to cython .pyx and .so files.
        """,
    ),
    cache_dir=BrianPreference(
        default=None,
        validator=lambda x: x is None or isinstance(x, str),
        docs="""
        Location of the cache directory for Cython files. By default,
        will be stored in a ``brian_extensions`` subdirectory
        where Cython inline stores its temporary files
        (the result of ``get_cython_cache_dir()``).
        """,
    ),
    delete_source_files=BrianPreference(
        default=True,
        docs="""
        Whether to delete source files after compiling. The Cython
        source files can take a significant amount of disk space, and
        are not used anymore when the compiled library file exists.
        They are therefore deleted by default, but keeping them around
        can be useful for debugging.
        """,
    ),
)


class CythonCodeObject(NumpyCodeObject):
    """
    Execute code using Cython.
    """

    templater = Templater(
        "brian2.codegen.runtime.cython_rt",
        ".pyx",
        env_globals={
            "cpp_dtype": get_cpp_dtype,
            "numpy_dtype": get_numpy_dtype,
            "dtype": numpy.dtype,
            "constant_or_scalar": constant_or_scalar,
        },
    )
    generator_class = CythonCodeGenerator
    class_name = "cython"

    def __init__(
        self,
        owner,
        code,
        variables,
        variable_indices,
        template_name,
        template_source,
        compiler_kwds,
        name="cython_code_object*",
    ):
        check_compiler_kwds(
            compiler_kwds,
            [
                "libraries",
                "include_dirs",
                "library_dirs",
                "runtime_library_dirs",
                "sources",
            ],
            "Cython",
        )
        super().__init__(
            owner,
            code,
            variables,
            variable_indices,
            template_name,
            template_source,
            compiler_kwds={},  # do not pass the actual args
            name=name,
        )
        self.compiler, self.extra_compile_args = get_compiler_and_args()
        self.define_macros = list(prefs["codegen.cpp.define_macros"])
        self.extra_link_args = list(prefs["codegen.cpp.extra_link_args"])
        self.headers = []  # not actually used

        self.include_dirs = list(prefs["codegen.cpp.include_dirs"]) + compiler_kwds.get(
            "include_dirs", []
        )
        self.include_dirs = list(prefs["codegen.cpp.include_dirs"])

        self.library_dirs = list(prefs["codegen.cpp.library_dirs"]) + compiler_kwds.get(
            "library_dirs", []
        )

        self.runtime_library_dirs = list(
            prefs["codegen.cpp.runtime_library_dirs"]
        ) + compiler_kwds.get("runtime_library_dirs", [])

        self.libraries = list(prefs["codegen.cpp.libraries"]) + compiler_kwds.get(
            "libraries", []
        )
        self.sources = compiler_kwds.get("sources", [])

    @classmethod
    def is_available(cls):
        try:
            compiler, extra_compile_args = get_compiler_and_args()
            code = """
            #cython: language_level=3
            def main():
                cdef int x
                x = 0"""
            compiled = cython_extension_manager.create_extension(
                code,
                compiler=compiler,
                extra_compile_args=extra_compile_args,
                extra_link_args=prefs["codegen.cpp.extra_link_args"],
                include_dirs=prefs["codegen.cpp.include_dirs"],
                library_dirs=prefs["codegen.cpp.library_dirs"],
                runtime_library_dirs=prefs["codegen.cpp.runtime_library_dirs"],
            )
            compiled.main()
            return True
        except Exception as ex:
            msg = (
                f"Cannot use Cython, a test compilation failed: {str(ex)} "
                f"({ex.__class__.__name__})"
            )
            if platform.system() != "Windows":
                msg += (
                    "\nCertain compiler configurations (e.g. clang in a conda "
                    "environment on OS X) are known to be problematic. Note that "
                    "you can switch the compiler by setting the 'CC' and 'CXX' "
                    "environment variables. For example, you may want to try "
                    "'CC=gcc' and 'CXX=g++'."
                )
            logger.warn(msg, "failed_compile_test")
            return False

    def compile_block(self, block):
        code = getattr(self.code, block, "").strip()
        if not code or "EMPTY_CODE_BLOCK" in code:
            return None
        return cython_extension_manager.create_extension(
            code,
            define_macros=self.define_macros,
            libraries=self.libraries,
            extra_compile_args=self.extra_compile_args,
            extra_link_args=self.extra_link_args,
            include_dirs=self.include_dirs,
            library_dirs=self.library_dirs,
            runtime_library_dirs=self.runtime_library_dirs,
            compiler=self.compiler,
            owner_name=f"{self.owner.name}_{self.template_name}",
            sources=self.sources,
        )

    def run_block(self, block):
        compiled_code = self.compiled_code[block]
        if compiled_code:
            try:
                return compiled_code.main(self.namespace)
            except Exception as exc:
                message = (
                    "An exception occured during the execution of the "
                    f"'{block}' block of code object '{self.name}'.\n"
                )
                raise BrianObjectException(message, self.owner) from exc

    def _insert_func_namespace(self, func):
        impl = func.implementations[self]
        func_namespace = impl.get_namespace(self.owner)
        if func_namespace is not None:
            self.namespace.update(func_namespace)
        if impl.dependencies is not None:
            for dep in impl.dependencies.values():
                self._insert_func_namespace(dep)

    def variables_to_namespace(self):
        # Variables can refer to values that are either constant (e.g. dt)
        # or change every timestep (e.g. t). We add the values of the
        # constant variables here and add the names of non-constant variables
        # to a list

        # A list containing tuples of name and a function giving the value
        self.nonconstant_values = []

        for name, var in self.variables.items():
            if isinstance(var, Function):
                self._insert_func_namespace(var)
            if isinstance(var, (AuxiliaryVariable, Subexpression)):
                continue
            try:
                value = var.get_value()
            except (TypeError, AttributeError):
                # A dummy Variable without value or a function
                self.namespace[name] = var
                continue

            if isinstance(var, ArrayVariable):
                self.namespace[self.device.get_array_name(var, self.variables)] = value
                self.namespace[f"_num{name}"] = var.get_len()
                if var.scalar and var.constant:
                    self.namespace[name] = value.item()
            else:
                self.namespace[name] = value

            if isinstance(var, DynamicArrayVariable):
                dyn_array_name = self.generator_class.get_array_name(
                    var, access_data=False
                )
                self.namespace[dyn_array_name] = self.device.get_value(
                    var, access_data=False
                )

            # Also provide the Variable object itself in the namespace (can be
            # necessary for resize operations, for example)
            self.namespace[f"_var_{name}"] = var

        # Get all identifiers in the code -- note that this is not a smart
        # function, it will get identifiers from strings, comments, etc. This
        # is not a problem here, since we only use this list to filter out
        # things. If we include something incorrectly, this only means that we
        # will pass something into the namespace unnecessarily.
        all_identifiers = get_identifiers(self.code.run)
        # Filter out all unneeded objects
        self.namespace = {
            k: v for k, v in self.namespace.items() if k in all_identifiers
        }

        # There is one type of objects that we have to inject into the
        # namespace with their current value at each time step: dynamic
        # arrays that change in size during runs, where the size change is not
        # initiated by the template itself
        for name, var in self.variables.items():
            if isinstance(var, DynamicArrayVariable) and var.needs_reference_update:
                array_name = self.device.get_array_name(var, self.variables)
                if array_name in self.namespace:
                    self.nonconstant_values.append((array_name, var.get_value))
                if f"_num{name}" in self.namespace:
                    self.nonconstant_values.append((f"_num{name}", var.get_len))

    def update_namespace(self):
        # update the values of the non-constant values in the namespace
        for name, func in self.nonconstant_values:
            self.namespace[name] = func()


codegen_targets.add(CythonCodeObject)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
STDP modulated with reward

Adapted from Fig. 1c of:

Eugene M. Izhikevich
Solving the distal reward problem through linkage of STDP and dopamine signaling.
Cerebral cortex 17, no. 10 (2007): 2443-2452.

Note:
The variable "mode" can switch the behavior of the synapse from "Classical STDP" to "Dopamine modulated STDP".

Author: Guillaume Dumas (Institut Pasteur)
Date: 2018-08-24
"""
from brian2 import *

# Parameters
simulation_duration = 6 * second

## Neurons
taum = 10*ms
Ee = 0*mV
vt = -54*mV
vr = -60*mV
El = -74*mV
taue = 5*ms

## STDP
taupre = 20*ms
taupost = taupre
gmax = .01
dApre = .01
dApost = -dApre * taupre / taupost * 1.05
dApost *= gmax
dApre *= gmax

## Dopamine signaling
tauc = 1000*ms
taud = 200*ms
taus = 1*ms
epsilon_dopa = 5e-3

# Setting the stage

## Stimuli section
input_indices = array([0, 1, 0, 1, 1, 0, 
                       0, 1, 0, 1, 1, 0])
input_times = array([ 500,  550, 1000, 1010, 1500, 1510, 
                     3500, 3550, 4000, 4010, 4500, 4510])*ms
spike_input = SpikeGeneratorGroup(2, input_indices, input_times)

neurons = NeuronGroup(2, '''dv/dt = (ge * (Ee-v) + El - v) / taum : volt
                            dge/dt = -ge / taue : 1''',
                      threshold='v>vt', reset='v = vr',
                      method='euler')
neurons.v = vr
neurons_monitor = SpikeMonitor(neurons)

synapse = Synapses(spike_input, neurons, 
                   model='''s: volt''',
                   on_pre='v += s')
synapse.connect(i=[0, 1], j=[0, 1])
synapse.s = 100. * mV

## STDP section
synapse_stdp = Synapses(neurons, neurons,
                   model='''mode: 1
                         dc/dt = -c / tauc : 1 (clock-driven)
                         dd/dt = -d / taud : 1 (clock-driven)
                         ds/dt = mode * c * d / taus : 1 (clock-driven)
                         dApre/dt = -Apre / taupre : 1 (event-driven)
                         dApost/dt = -Apost / taupost : 1 (event-driven)''',
                   on_pre='''ge += s
                          Apre += dApre
                          c = clip(c + mode * Apost, -gmax, gmax)
                          s = clip(s + (1-mode) * Apost, -gmax, gmax)
                          ''',
                   on_post='''Apost += dApost
                          c = clip(c + mode * Apre, -gmax, gmax)
                          s = clip(s + (1-mode) * Apre, -gmax, gmax)
                          ''',
                   method='euler'
                   )
synapse_stdp.connect(i=0, j=1)
synapse_stdp.mode = 0
synapse_stdp.s = 1e-10
synapse_stdp.c = 1e-10
synapse_stdp.d = 0
synapse_stdp_monitor = StateMonitor(synapse_stdp, ['s', 'c', 'd'], record=[0])

## Dopamine signaling section
dopamine_indices = array([0, 0, 0])
dopamine_times = array([3520, 4020, 4520])*ms
dopamine = SpikeGeneratorGroup(1, dopamine_indices, dopamine_times)
dopamine_monitor = SpikeMonitor(dopamine)
reward = Synapses(dopamine, synapse_stdp, model='''''',
                            on_pre='''d_post += epsilon_dopa''',
                            method='exact')
reward.connect()

# Simulation
## Classical STDP
synapse_stdp.mode = 0
run(simulation_duration/2)
## Dopamine modulated STDP
synapse_stdp.mode = 1
run(simulation_duration/2)

# Visualisation
dopamine_indices, dopamine_times = dopamine_monitor.it
neurons_indices, neurons_times = neurons_monitor.it
figure(figsize=(12, 6))
subplot(411)
plot([0.05, 2.95], [2.7, 2.7], linewidth=5, color='k')
text(1.5, 3, 'Classical STDP', horizontalalignment='center', fontsize=20)
plot([3.05, 5.95], [2.7, 2.7], linewidth=5, color='k')
text(4.5, 3, 'Dopamine modulated STDP', horizontalalignment='center', fontsize=20)
plot(neurons_times, neurons_indices, 'ob')
plot(dopamine_times, dopamine_indices + 2, 'or')
xlim([0, simulation_duration/second])
ylim([-0.5, 4])
yticks([0, 1, 2], ['Pre-neuron', 'Post-neuron', 'Reward'])
xticks([])
subplot(412)
plot(synapse_stdp_monitor.t/second, synapse_stdp_monitor.d.T/gmax, 'r-')
xlim([0, simulation_duration/second])
ylabel('Extracellular\ndopamine d(t)')
xticks([])
subplot(413)
plot(synapse_stdp_monitor.t/second, synapse_stdp_monitor.c.T/gmax, 'b-')
xlim([0, simulation_duration/second])
ylabel('Eligibility\ntrace c(t)')
xticks([])
subplot(414)
plot(synapse_stdp_monitor.t/second, synapse_stdp_monitor.s.T/gmax, 'g-')
xlim([0, simulation_duration/second])
ylabel('Synaptic\nstrength s(t)')
xlabel('Time (s)')
tight_layout()
show()



{# USES_VARIABLES { N, _clock_t, count,
                    _source_start, _source_stop} #}
{% extends 'common_group.pyx' %}

{% block maincode %}

    {#  Get the name of the array that stores these events (e.g. the spikespace array) #}
    {% set _eventspace = get_array_name(eventspace_variable) %}

    cdef size_t _num_events = {{_eventspace}}[_num{{_eventspace}}-1]
    cdef size_t _start_idx, _end_idx, _curlen, _newlen, _j
    {% for varname, var in record_variables | dictsort %}
    cdef {{cpp_dtype(var.dtype)}}[:] _{{varname}}_view
    {% endfor %}
    if _num_events > 0:
        # For subgroups, we do not want to record all spikes
        # We assume that spikes are ordered
        _start_idx = _num_events
        _end_idx = _num_events
        for _j in range(_num_events):
            _idx = {{_eventspace}}[_j]
            if _idx >= _source_start:
                _start_idx = _j
                break
        for _j in range(_num_events-1, _start_idx-1, -1):
            _idx = {{_eventspace}}[_j]
            if _idx < _source_stop:
                break
            _end_idx = _j
        _num_events = _end_idx - _start_idx
        if _num_events > 0:
            # scalar code
            _vectorisation_idx = 1
            {{ scalar_code|autoindent }}
            _curlen = {{N}}
            _newlen = _curlen + _num_events
            # Resize the arrays
            _owner.resize(_newlen)
            {{N}} = _newlen
            {% for varname, var in record_variables | dictsort %}
            _{{varname}}_view = {{get_array_name(var, access_data=False)}}.data
            {% endfor %}
            # Copy the values across
            for _j in range(_start_idx, _end_idx):
                _idx = {{_eventspace}}[_j]
                _vectorisation_idx = _idx
                {{ vector_code|autoindent }}
                {% for varname in record_variables | sort %}
                _{{varname}}_view [_curlen + _j - _start_idx] = _to_record_{{varname}}
                {% endfor %}
                {{count}}[_idx - _source_start] += 1
{% endblock %}



# encoding: utf8
"""
Burst generation in the Hinsmarsh-Rose model. Reproduces Figure 6 of:

Hindmarsh, J. L., and R. M. Rose.
“A Model of Neuronal Bursting Using Three Coupled First Order Differential Equations.”
Proceedings of the Royal Society of London. Series B, Biological Sciences 221, no. 1222 (1984): 87–102.
"""
from brian2 import *

# In the original model, time is measured in arbitrary time units
time_unit = 1*ms
defaultclock.dt = time_unit/10

x_1 = -1.6  # leftmost equilibrium point of the model without adaptation
a = 1; b = 3; c = 1; d = 5
r = 0.001; s = 4
eqs = '''
dx/dt = (y - a*x**3 + b*x**2 + I - z)/time_unit : 1
dy/dt = (c - d*x**2 - y)/time_unit : 1
dz/dt = r*(s*(x - x_1) - z)/time_unit : 1
I : 1 (constant)
'''

# We run the model with three different currents
neuron = NeuronGroup(3, eqs, method='rk4')

# Set all variables to their equilibrium point
neuron.x = x_1
neuron.y = 'c - d*x**2'
neuron.z = 'r*(s*(x - x_1))'

# Set the constant current input
neuron.I = [0.4, 2, 4]

# Record the "membrane potential"
mon = StateMonitor(neuron, 'x', record=True)

run(2100*time_unit)

ax_top = plt.subplot2grid((2, 3), (0, 0), colspan=3)
ax_bottom_l = plt.subplot2grid((2, 3), (1, 0), colspan=2)
ax_bottom_r = plt.subplot2grid((2, 3), (1, 2))
for ax in [ax_top, ax_bottom_l, ax_bottom_r]:
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.set(ylim=(-2, 2), yticks=[-2, 0, 2])

ax_top.plot(mon.t/time_unit, mon.x[0])

ax_bottom_l.plot(mon.t/time_unit, mon.x[1])
ax_bottom_l.set_xlim(700, 2100)

ax_bottom_r.plot(mon.t/time_unit, mon.x[2])
ax_bottom_r.set_xlim(1400, 2100)
ax_bottom_r.set_yticks([])

plt.show()



"""
Classes for providing external input to a network.
"""

from .binomial import *
from .poissongroup import *
from .poissoninput import *
from .spikegeneratorgroup import *
from .timedarray import *

__all__ = [
    "BinomialFunction",
    "PoissonGroup",
    "PoissonInput",
    "SpikeGeneratorGroup",
    "TimedArray",
]



import pytest

from brian2 import *
from brian2.codegen.templates import Templater


@pytest.mark.codegen_independent
def test_templates():
    T1 = Templater(
        "brian2.tests.test_templates.fake_package_1",
        env_globals={"f": lambda: "F1", "g": lambda: "G1"},
        extension=".txt",
    )
    T2 = T1.derive(
        "brian2.tests.test_templates.fake_package_2", env_globals={"f": lambda: "F2"}
    )
    ns = {}
    for i, T in enumerate([T1, T2], start=1):
        for c in ["A", "B", "C", "D"]:
            ns[c + str(i)] = getattr(T, c)("", "")
    # for k, v in ns.items():
    #     print k, v
    assert "A1" in ns["A1"]
    assert "B1" in ns["A1"]
    assert "F1" in ns["A1"]
    assert "G1" in ns["A1"]
    assert "A2" in ns["A2"]
    assert "F2" in ns["A2"]
    assert "G1" in ns["A2"]
    assert "B1" not in ns["A2"]
    assert "B1" in ns["B1"]
    assert "B1" in ns["B2"]
    assert "C1" in ns["C1"]
    assert "D1" in ns["C1"]
    assert "C1" in ns["C2"]
    assert "D2" in ns["C2"]
    assert "D1" not in ns["C2"]
    assert "D1" in ns["D1"]
    assert "D2" in ns["D2"]


if __name__ == "__main__":
    test_templates()



"""
This module defines the `StateUpdateMethod` class that acts as a base class for
all stateupdaters and allows to register stateupdaters so that it is able to
return a suitable stateupdater object for a given set of equations. This is used
for example in `NeuronGroup` when no state updater is given explicitly.
"""

import time
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable

from brian2.utils.caching import cached
from brian2.utils.logger import get_logger

__all__ = ["StateUpdateMethod"]

logger = get_logger(__name__)


class UnsupportedEquationsException(Exception):
    pass


def extract_method_options(method_options, default_options):
    """
    Helper function to check ``method_options`` against options understood by
    this state updater, and setting default values for all unspecified options.

    Parameters
    ----------
    method_options : dict or None
        The options that the user specified for the state update.
    default_options : dict
        The default option values for this state updater (each admissible option
        needs to be present in this dictionary). To specify that a state updater
        does not take any options, provide an empty dictionary as the argument.

    Returns
    -------
    options : dict
        The final dictionary with all the options either at their default or at
        the user-specified value.

    Raises
    ------
    KeyError
        If the user specifies an option that is not understood by this state
        updater.

    Examples
    --------
    >>> options = extract_method_options({'a': True}, default_options={'b': False, 'c': False})  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    KeyError: 'method_options specifies "a", but this is not an option for this state updater. Avalaible options are: "b", "c".'
    >>> options = extract_method_options({'a': True}, default_options={})  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    KeyError: 'method_options specifies "a", but this is not an option for this state updater. This state updater does not accept any options.'
    >>> options = extract_method_options({'a': True}, default_options={'a': False, 'b': False})
    >>> sorted(options.items())
    [('a', True), ('b', False)]
    """
    if method_options is None:
        method_options = {}
    for key in method_options:
        if key not in default_options:
            if len(default_options):
                keys = sorted(default_options.keys())
                options = (
                    "Available options are: "
                    + ", ".join(f"'{key}'" for key in keys)
                    + "."
                )
            else:
                options = "This state updater does not accept any options."
            raise KeyError(
                f"method_options specifies '{key}', but this "
                "is not an option for this state updater. "
                f"{options}"
            )
    filled_options = dict(default_options)
    filled_options.update(method_options)
    return filled_options


class StateUpdateMethod(metaclass=ABCMeta):
    stateupdaters = dict()

    @abstractmethod
    def __call__(self, equations, variables=None, method_options=None):
        """
        Generate abstract code from equations. The method also gets the
        the variables because some state updaters have to check whether
        variable names reflect other state variables (which can change from
        timestep to timestep) or are external values (which stay constant during
        a run)  For convenience, this arguments are optional -- this allows to
        directly see what code a state updater generates for a set of equations
        by simply writing ``euler(eqs)``, for example.

        Parameters
        ----------
        equations : `Equations`
            The model equations.
        variables : dict, optional
            The `Variable` objects for the model variables.
        method_options : dict, optional
            Additional options specific to the state updater.
        Returns
        -------
        code : str
            The abstract code performing a state update step.
        """
        pass

    @staticmethod
    def register(name, stateupdater):
        """
        Register a state updater. Registered state updaters can be referred to
        via their name.

        Parameters
        ----------
        name : str
            A short name for the state updater (e.g. `'euler'`)
        stateupdater : `StateUpdaterMethod`
            The state updater object, e.g. an `ExplicitStateUpdater`.
        """

        # only deal with lower case names -- we don't want to have 'Euler' and
        # 'euler', for example
        name = name.lower()
        if name in StateUpdateMethod.stateupdaters:
            raise ValueError(
                f"A stateupdater with the name '{name}' has already been registered"
            )

        if not isinstance(stateupdater, StateUpdateMethod):
            raise ValueError(
                f"Given stateupdater of type {type(stateupdater)} does "
                "not seem to be a valid stateupdater."
            )

        StateUpdateMethod.stateupdaters[name] = stateupdater

    @staticmethod
    @cached
    def apply_stateupdater(
        equations, variables, method, method_options=None, group_name=None
    ):
        """
        apply_stateupdater(equations, variables, method, method_options=None, group_name=None)

        Applies a given state updater to equations. If a `method` is given, the
        state updater with the given name is used or if is a callable, then it
        is used directly. If a `method` is a list of names, all the
        methods will be tried until one that doesn't raise an
        `UnsupportedEquationsException` is found.

        Parameters
        ----------
        equations : `Equations`
            The model equations.
        variables : `dict`
            The dictionary of `Variable` objects, describing the internal
            model variables.
        method : {callable, str, list of str}
            A callable usable as a state updater, the name of a registered
            state updater or a list of names of state updaters.

        Returns
        -------
        abstract_code : str
            The code integrating the given equations.
        """
        if isinstance(method, Iterable) and not isinstance(method, str):
            the_method = None
            start_time = time.time()
            for one_method in method:
                try:
                    one_method_start_time = time.time()
                    code = StateUpdateMethod.apply_stateupdater(
                        equations, variables, one_method, group_name=group_name
                    )
                    the_method = one_method
                    one_method_time = time.time() - one_method_start_time
                    break
                except UnsupportedEquationsException:
                    pass
                except TypeError:
                    raise TypeError(
                        "Each element in the list of methods has "
                        "to be a string or a callable, got "
                        f"{type(one_method)}."
                    )
            total_time = time.time() - start_time
            if the_method is None:
                raise ValueError(
                    "No stateupdater that is suitable for the "
                    "given equations has been found."
                )

            # If only one method was tried
            if method[0] == the_method:
                timing = f"took {one_method_time:.2f}s"
            else:
                timing = (
                    f"took {one_method_time:.2f}s, trying other methods took "
                    f"{total_time - one_method_time:.2f}s"
                )

            if group_name is not None:
                msg_text = (
                    "No numerical integration method specified for group "
                    f"'{group_name}', using method '{the_method}' ({timing})."
                )
            else:
                msg_text = (
                    "No numerical integration method specified, "
                    f"using method '{the_method}' ({timing})."
                )
            logger.info(msg_text, "method_choice")
        else:
            if callable(method):
                # if this is a standard state updater, i.e. if it has a
                # can_integrate method, check this method and raise a warning if it
                # claims not to be applicable.
                stateupdater = method
                method = getattr(
                    stateupdater, "__name__", repr(stateupdater)
                )  # For logging, get a nicer name
            elif isinstance(method, str):
                method = method.lower()  # normalize name to lower case
                stateupdater = StateUpdateMethod.stateupdaters.get(method, None)
                if stateupdater is None:
                    raise ValueError(
                        "No state updater with the name '{method}' is known."
                    )
            else:
                raise TypeError(
                    "method argument has to be a string, a "
                    "callable, or an iterable of such objects. "
                    f"Got {type(method)}"
                )
            start_time = time.time()
            code = stateupdater(equations, variables, method_options)
            method_time = time.time() - start_time
            timing = "took %.2fs" % method_time
            if group_name is not None:
                logger.debug(
                    f"Group {group_name}: using numerical integration "
                    f"method {method} ({timing})",
                    "method_choice",
                )
            else:
                logger.debug(
                    f"Using numerical integration method: {method} f({{timing}})",
                    "method_choice",
                )

        return code



"""
AST parsing based analysis of expressions
"""

import ast

from brian2.core.functions import Function
from brian2.parsing.rendering import NodeRenderer
from brian2.units.fundamentalunits import (
    DIMENSIONLESS,
    DimensionMismatchError,
    Unit,
    get_dimensions,
    get_unit_for_display,
    have_same_dimensions,
)

__all__ = ["parse_expression_dimensions"]


def is_boolean_expression(expr, variables):
    """
    Determines if an expression is of boolean type or not

    Parameters
    ----------

    expr : str
        The expression to test
    variables : dict-like of `Variable`
        The variables used in the expression.

    Returns
    -------
    isbool : bool
        Whether or not the expression is boolean.

    Raises
    ------
    SyntaxError
        If the expression ought to be boolean but is not,
        for example ``x<y and z`` where ``z`` is not a boolean variable.

    Notes
    -----
    We test the following cases recursively on the abstract syntax tree:

    * The node is a boolean operation. If all the subnodes are boolean
      expressions we return ``True``, otherwise we raise the ``SyntaxError``.
    * The node is a function call, we return ``True`` or ``False`` depending
      on whether the function description has the ``_returns_bool`` attribute
      set.
    * The node is a variable name, we return ``True`` or ``False`` depending
      on whether ``is_boolean`` attribute is set or if the name is ``True`` or
      ``False``.
    * The node is a comparison, we return ``True``.
    * The node is a unary operation, we return ``True`` if the operation is
      ``not``, otherwise ``False``.
    * Otherwise we return ``False``.
    """

    # If we are working on a string, convert to the top level node
    if isinstance(expr, str):
        mod = ast.parse(expr, mode="eval")
        expr = mod.body

    if expr.__class__ is ast.BoolOp:
        if all(is_boolean_expression(node, variables) for node in expr.values):
            return True
        else:
            raise SyntaxError(
                "Expression ought to be boolean but is not (e.g. 'x<y and 3')"
            )
    elif expr.__class__ is ast.Constant:
        value = expr.value
        if value is True or value is False:
            return True
        elif value is None:
            raise ValueError("Do not know how to deal with 'None'")
    elif expr.__class__ is ast.Name:
        name = expr.id
        if name in variables:
            return variables[name].is_boolean
        else:
            return name == "True" or name == "False"
    elif expr.__class__ is ast.Call:
        name = expr.func.id
        if name in variables and hasattr(variables[name], "_returns_bool"):
            return variables[name]._returns_bool
        else:
            raise SyntaxError(f"Unknown function {name}")
    elif expr.__class__ is ast.Compare:
        return True
    elif expr.__class__ is ast.UnaryOp:
        return expr.op.__class__.__name__ == "Not"
    else:
        return False


def _get_value_from_expression(expr, variables):
    """
    Returns the scalar value of an expression, and checks its validity.

    Parameters
    ----------
    expr : str or `ast.Expression`
        The expression to check.
    variables : dict of `Variable` objects
        The information about all variables used in `expr` (including `Constant`
        objects for external variables)

    Returns
    -------
    value : float
        The value of the expression

    Raises
    ------
    SyntaxError
        If the expression cannot be evaluated to a scalar value
    DimensionMismatchError
        If any part of the expression is dimensionally inconsistent.
    """
    # If we are working on a string, convert to the top level node
    if isinstance(expr, str):
        mod = ast.parse(expr, mode="eval")
        expr = mod.body

    if expr.__class__ is ast.Name:
        name = expr.id
        if name in variables:
            if not getattr(variables[name], "constant", False):
                raise SyntaxError(f"Value {name} is not constant")
            if not getattr(variables[name], "scalar", False):
                raise SyntaxError(f"Value {name} is not scalar")
            return variables[name].get_value()
        elif name in ["True", "False"]:
            return 1.0 if name == "True" else 0.0
        else:
            raise ValueError(f"Unknown identifier {name}")
    elif expr.__class__ is ast.Constant:
        return expr.value
    elif expr.__class__ is ast.BoolOp:
        raise SyntaxError(
            "Cannot determine the numerical value for a boolean operation."
        )
    elif expr.__class__ is ast.Compare:
        raise SyntaxError(
            "Cannot determine the numerical value for a boolean operation."
        )
    elif expr.__class__ is ast.Call:
        raise SyntaxError("Cannot determine the numerical value for a function call.")
    elif expr.__class__ is ast.BinOp:
        op = expr.op.__class__.__name__
        left = _get_value_from_expression(expr.left, variables)
        right = _get_value_from_expression(expr.right, variables)
        if op == "Add" or op == "Sub":
            v = left + right
        elif op == "Mult":
            v = left * right
        elif op == "Div":
            v = float(left) / right
        elif op == "FloorDiv":
            v = left // right
        elif op == "Pow":
            v = left**right
        elif op == "Mod":
            v = left % right
        else:
            raise SyntaxError(f"Unsupported operation {op}")
        return v
    elif expr.__class__ is ast.UnaryOp:
        op = expr.op.__class__.__name__
        # check validity of operand and get its unit
        v = _get_value_from_expression(expr.operand, variables)
        if op == "Not":
            raise SyntaxError(
                "Cannot determine the numerical value for a boolean operation."
            )
        if op == "USub":
            return -v
        else:
            raise SyntaxError(f"Unknown unary operation {op}")
    else:
        raise SyntaxError(f"Unsupported operation {str(expr.__class__)}")


def parse_expression_dimensions(expr, variables, orig_expr=None):
    """
    Returns the unit value of an expression, and checks its validity

    Parameters
    ----------
    expr : str
        The expression to check.
    variables : dict
        Dictionary of all variables used in the `expr` (including `Constant`
        objects for external variables)

    Returns
    -------
    unit : Quantity
        The output unit of the expression

    Raises
    ------
    SyntaxError
        If the expression cannot be parsed, or if it uses ``a**b`` for ``b``
        anything other than a constant number.
    DimensionMismatchError
        If any part of the expression is dimensionally inconsistent.
    """

    # If we are working on a string, convert to the top level node
    if isinstance(expr, str):
        orig_expr = expr
        mod = ast.parse(expr, mode="eval")
        expr = mod.body
    if expr.__class__ is ast.Name:
        name = expr.id
        # Raise an error if a function is called as if it were a variable
        # (most of the time this happens for a TimedArray)
        if name in variables and isinstance(variables[name], Function):
            raise SyntaxError(
                f"{name} was used like a variable/constant, but it is a function.",
                ("<string>", expr.lineno, expr.col_offset + 1, orig_expr),
            )
        if name in variables:
            return get_dimensions(variables[name])
        elif name in ["True", "False"]:
            return DIMENSIONLESS
        else:
            raise KeyError(f"Unknown identifier {name}")
    elif expr.__class__ is ast.Constant:
        return DIMENSIONLESS
    elif expr.__class__ is ast.BoolOp:
        # check that the units are valid in each subexpression
        for node in expr.values:
            parse_expression_dimensions(node, variables, orig_expr=orig_expr)
        # but the result is a bool, so we just return 1 as the unit
        return DIMENSIONLESS
    elif expr.__class__ is ast.Compare:
        # check that the units are consistent in each subexpression
        subexprs = [expr.left] + expr.comparators
        subunits = []
        for node in subexprs:
            subunits.append(
                parse_expression_dimensions(node, variables, orig_expr=orig_expr)
            )
        for left_dim, right_dim in zip(subunits[:-1], subunits[1:]):
            if not have_same_dimensions(left_dim, right_dim):
                left_expr = NodeRenderer().render_node(expr.left)
                right_expr = NodeRenderer().render_node(expr.comparators[0])
                dim_left = get_dimensions(left_dim)
                dim_right = get_dimensions(right_dim)
                msg = (
                    "Comparison of expressions with different units. Expression "
                    f"'{left_expr}' has unit ({dim_left}), while expression "
                    f"'{right_expr}' has units ({dim_right})."
                )
                raise DimensionMismatchError(msg)
        # but the result is a bool, so we just return 1 as the unit
        return DIMENSIONLESS
    elif expr.__class__ is ast.Call:
        if len(expr.keywords):
            raise ValueError("Keyword arguments not supported.")
        elif getattr(expr, "starargs", None) is not None:
            raise ValueError("Variable number of arguments not supported")
        elif getattr(expr, "kwargs", None) is not None:
            raise ValueError("Keyword arguments not supported")

        func = variables.get(expr.func.id, None)
        if func is None:
            raise SyntaxError(
                f"Unknown function {expr.func.id}",
                ("<string>", expr.lineno, expr.col_offset + 1, orig_expr),
            )
        if not hasattr(func, "_arg_units") or not hasattr(func, "_return_unit"):
            raise ValueError(
                f"Function {expr.func.id} does not specify how it deals with units."
            )

        if len(func._arg_units) != len(expr.args):
            raise SyntaxError(
                f"Function '{expr.func.id}' was called with "
                f"{len(expr.args)} parameters, needs "
                f"{len(func._arg_units)}.",
                (
                    "<string>",
                    expr.lineno,
                    expr.col_offset + len(expr.func.id) + 1,
                    orig_expr,
                ),
            )

        for idx, (arg, expected_unit) in enumerate(zip(expr.args, func._arg_units)):
            arg_unit = parse_expression_dimensions(arg, variables, orig_expr=orig_expr)
            # A "None" in func._arg_units means: No matter what unit
            if expected_unit is None:
                continue
            # A string means: same unit as other argument
            elif isinstance(expected_unit, str):
                arg_idx = func._arg_names.index(expected_unit)
                expected_unit = parse_expression_dimensions(
                    expr.args[arg_idx], variables, orig_expr=orig_expr
                )
                if not have_same_dimensions(arg_unit, expected_unit):
                    msg = (
                        f"Argument number {idx + 1} for function "
                        f"{expr.func.id} was supposed to have the "
                        f"same units as argument number {arg_idx + 1}, but "
                        f"'{NodeRenderer().render_node(arg)}' has unit "
                        f"{get_unit_for_display(arg_unit)}, while "
                        f"'{NodeRenderer().render_node(expr.args[arg_idx])}' "
                        f"has unit {get_unit_for_display(expected_unit)}"
                    )
                    raise DimensionMismatchError(msg)
            elif expected_unit == bool:
                if not is_boolean_expression(arg, variables):
                    rendered_arg = NodeRenderer().render_node(arg)
                    raise TypeError(
                        f"Argument number {idx + 1} for function "
                        f"'{expr.func.id}' was expected to be a boolean "
                        f"value, but is '{rendered_arg}'."
                    )
            else:
                if not have_same_dimensions(arg_unit, expected_unit):
                    rendered_arg = NodeRenderer().render_node(arg)
                    arg_unit_dim = get_dimensions(arg_unit)
                    expected_unit_dim = get_dimensions(expected_unit)
                    msg = (
                        f"Argument number {idx+1} for function {expr.func.id} does "
                        f"not have the correct units. Expression '{rendered_arg}' "
                        f"has units ({arg_unit_dim}), but "
                        "should be "
                        f"({expected_unit_dim})."
                    )
                    raise DimensionMismatchError(msg)

        if func._return_unit == bool:
            return DIMENSIONLESS
        elif isinstance(func._return_unit, (Unit, int)):
            # Function always returns the same unit
            return getattr(func._return_unit, "dim", DIMENSIONLESS)
        else:
            # Function returns a unit that depends on the arguments
            arg_units = [
                parse_expression_dimensions(arg, variables, orig_expr=orig_expr)
                for arg in expr.args
            ]
            return func._return_unit(*arg_units).dim

    elif expr.__class__ is ast.BinOp:
        op = expr.op.__class__.__name__
        left_dim = parse_expression_dimensions(
            expr.left, variables, orig_expr=orig_expr
        )
        right_dim = parse_expression_dimensions(
            expr.right, variables, orig_expr=orig_expr
        )
        if op in ["Add", "Sub", "Mod"]:
            # dimensions should be the same
            if left_dim is not right_dim:
                op_symbol = {"Add": "+", "Sub": "-", "Mod": "%"}.get(op)
                left_str = NodeRenderer().render_node(expr.left)
                right_str = NodeRenderer().render_node(expr.right)
                left_unit = get_unit_for_display(left_dim)
                right_unit = get_unit_for_display(right_dim)
                error_msg = (
                    f"Expression '{left_str} {op_symbol} {right_str}' uses "
                    f"inconsistent units ('{left_str}' has unit "
                    f"{left_unit}; '{right_str}' "
                    f"has unit {right_unit})."
                )
                raise DimensionMismatchError(error_msg)
            u = left_dim
        elif op == "Mult":
            u = left_dim * right_dim
        elif op == "Div":
            u = left_dim / right_dim
        elif op == "FloorDiv":
            if not (left_dim is DIMENSIONLESS and right_dim is DIMENSIONLESS):
                if left_dim is DIMENSIONLESS:
                    col_offset = expr.right.col_offset + 1
                else:
                    col_offset = expr.left.col_offset + 1
                raise SyntaxError(
                    "Floor division can only be used on dimensionless values.",
                    ("<string>", expr.lineno, col_offset, orig_expr),
                )
            u = DIMENSIONLESS
        elif op == "Pow":
            if left_dim is DIMENSIONLESS and right_dim is DIMENSIONLESS:
                return DIMENSIONLESS
            n = _get_value_from_expression(expr.right, variables)
            u = left_dim**n
        else:
            raise SyntaxError(
                f"Unsupported operation {op}",
                (
                    "<string>",
                    expr.lineno,
                    getattr(
                        expr.left,
                        "end_col_offset",
                        len(NodeRenderer().render_node(expr.left)),
                    )
                    + 1,
                    orig_expr,
                ),
            )
        return u
    elif expr.__class__ is ast.UnaryOp:
        op = expr.op.__class__.__name__
        # check validity of operand and get its unit
        u = parse_expression_dimensions(expr.operand, variables, orig_expr=orig_expr)
        if op == "Not":
            return DIMENSIONLESS
        else:
            return u
    else:
        raise SyntaxError(
            f"Unsupported operation {str(expr.__class__.__name__)}",
            ("<string>", expr.lineno, expr.col_offset + 1, orig_expr),
        )



"""
Module providing the `Synapses` class and related helper classes/functions.
"""

import functools
import numbers
import re
import weakref
from collections import defaultdict
from collections.abc import Mapping, MutableMapping, Sequence

import numpy as np

from brian2.codegen.codeobject import create_runner_codeobj
from brian2.codegen.translation import get_identifiers_recursively
from brian2.core.base import device_override, weakproxy_with_fallback
from brian2.core.namespace import get_local_namespace
from brian2.core.spikesource import SpikeSource
from brian2.core.variables import DynamicArrayVariable, Variables
from brian2.devices.device import get_device
from brian2.equations.equations import (
    DIFFERENTIAL_EQUATION,
    PARAMETER,
    SUBEXPRESSION,
    EquationError,
    Equations,
    check_subexpressions,
)
from brian2.groups.group import CodeRunner, Group, get_dtype
from brian2.groups.neurongroup import (
    SubexpressionUpdater,
    check_identifier_pre_post,
    extract_constant_subexpressions,
)
from brian2.parsing.bast import brian_ast
from brian2.parsing.expressions import (
    is_boolean_expression,
    parse_expression_dimensions,
)
from brian2.parsing.rendering import NodeRenderer
from brian2.stateupdaters.base import StateUpdateMethod, UnsupportedEquationsException
from brian2.stateupdaters.exact import linear
from brian2.synapses.parse_synaptic_generator_syntax import parse_synapse_generator
from brian2.units.allunits import second
from brian2.units.fundamentalunits import (
    DIMENSIONLESS,
    DimensionMismatchError,
    Quantity,
    fail_for_dimension_mismatch,
)
from brian2.utils.arrays import calc_repeats
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import get_identifiers, word_substitute

MAX_SYNAPSES = 2147483647

__all__ = ["Synapses"]

logger = get_logger(__name__)


class StateUpdater(CodeRunner):
    """
    The `CodeRunner` that updates the state variables of a `Synapses`
    at every timestep.
    """

    def __init__(self, group, method, clock, order, method_options=None):
        self.method_choice = method
        self.method_options = method_options
        CodeRunner.__init__(
            self,
            group,
            "stateupdate",
            clock=clock,
            when="groups",
            order=order,
            name=group.name + "_stateupdater",
            check_units=False,
            generate_empty_code=False,
        )

    def update_abstract_code(self, run_namespace):
        if len(self.group.equations) > 0:
            # Resolve variables in the equations to correctly perform checks
            # for repeated stateful functions (e.g. rand() calls)
            names = self.group.equations.names
            external_names = self.group.equations.identifiers | {"dt"}

            variables = self.group.resolve_all(
                names | external_names,
                run_namespace,
                # we don't need to raise any warnings
                # for the user here, warnings will
                # be raised in create_runner_codeobj
                user_identifiers=set(),
            )
            stateupdate_output = StateUpdateMethod.apply_stateupdater(
                self.group.equations,
                variables,
                self.method_choice,
                method_options=self.method_options,
                group_name=self.group.name,
            )
            if isinstance(stateupdate_output, str):
                self.abstract_code = stateupdate_output
            else:
                # Note that the reason to send self along with this method is so the StateUpdater
                # can be modified! i.e. in GSL StateUpdateMethod a custom CodeObject gets added
                # to the StateUpdater together with some auxiliary information
                self.abstract_code = stateupdate_output(self)
        else:
            self.abstract_code = ""


class SummedVariableUpdater(CodeRunner):
    """
    The `CodeRunner` that updates a value in the target group with the
    sum over values in the `Synapses` object.
    """

    def __init__(
        self, expression, target_varname, synapses, target, target_size_name, index_var
    ):
        # Handling sumped variables using the standard mechanisms is not
        # possible, we therefore also directly give the names of the arrays
        # to the template.

        code = f"""
        _synaptic_var = {expression}
        """
        self.target_varname = target_varname
        self.expression = expression
        self.target_var = synapses.variables[target_varname]
        self.target = target
        template_kwds = {
            "_target_var": self.target_var,
            "_target_size_name": target_size_name,
            "_index_var": synapses.variables[index_var],
            "_target_start": getattr(target, "start", 0),
            "_target_stop": getattr(target, "stop", -1),
        }

        CodeRunner.__init__(
            self,
            group=synapses,
            template="summed_variable",
            code=code,
            needed_variables=[target_varname, target_size_name, index_var],
            # We want to update the summed variable before
            # the target group gets updated
            clock=target.clock,
            when="groups",
            order=target.order - 1,
            name=synapses.name + "_summed_variable_" + target_varname,
            template_kwds=template_kwds,
        )

    def before_run(self, run_namespace):
        variables = self.group.resolve_all(self.expression.identifiers, run_namespace)
        rhs_unit = parse_expression_dimensions(self.expression.code, variables)
        fail_for_dimension_mismatch(
            self.target_var,
            # Using a quantity instead of dimensions
            # here makes fail_for_dimension_mismatch
            # state the dimensions as part of the error
            # message
            Quantity(1, dim=rhs_unit),
            "The target variable "
            f"'{self.target_varname}' does not have "
            "the same dimensions as the right-hand "
            f"side expression '{self.expression}'.",
        )
        super().before_run(run_namespace)


class SynapticPathway(CodeRunner, Group):
    """
    The `CodeRunner` that applies the pre/post statement(s) to the state
    variables of synapses where the pre-/postsynaptic group spiked in this
    time step.

    Parameters
    ----------

    synapses : `Synapses`
        Reference to the main `Synapses` object
    prepost : {'pre', 'post'}
        Whether this object should react to pre- or postsynaptic spikes
    objname : str, optional
        The name to use for the object, will be appendend to the name of
        `synapses` to create a name in the sense of `Nameable`. If ``None``
        is provided (the default), ``prepost`` will be used.
    delay : `Quantity`, optional
        A scalar delay (same delay for all synapses) for this pathway. If
        not given, delays are expected to vary between synapses.
    """

    def __init__(
        self, synapses, code, prepost, objname=None, delay=None, event="spike"
    ):
        self.code = code
        self.prepost = prepost
        self.event = event
        if prepost == "pre":
            self.source = synapses.source
            self.target = synapses.target
            self.synapse_sources = synapses.variables["_synaptic_pre"]
            self.synapse_targets = synapses.variables["_synaptic_post"]
            order = -1
        elif prepost == "post":
            self.source = synapses.target
            self.target = synapses.source
            self.synapse_sources = synapses.variables["_synaptic_post"]
            self.synapse_targets = synapses.variables["_synaptic_pre"]
            order = 1
        else:
            raise ValueError("prepost argument has to be either 'pre' or 'post'")
        self.synapses = weakref.proxy(synapses)
        # Allow to use the same indexing of the delay variable  as in the parent
        # Synapses object (e.g. 2d indexing with pre- and post-synaptic indices)
        self._indices = self.synapses._indices

        if objname is None:
            objname = prepost

        CodeRunner.__init__(
            self,
            synapses,
            "synapses",
            code=code,
            clock=self.source.clock,
            when="synapses",
            order=order,
            name=synapses.name + "_" + objname,
            template_kwds={"pathway": self},
        )

        self._pushspikes_codeobj = None

        self.spikes_start = self.source.start
        self.spikes_stop = self.source.stop
        self.eventspace_name = f"_{event}space"
        self.eventspace = None  # will be set in before_run
        # Setting the Synapses object instead of "self" as an owner makes
        # indexing conflicts disappear (e.g. with synapses connecting subgroups)
        self.variables = Variables(synapses)
        self.variables.add_reference(self.eventspace_name, self.source)
        self.variables.add_reference("N", synapses)
        if prepost == "pre":
            self.variables.add_reference("_n_sources", synapses, "N_pre")
            self.variables.add_reference("_n_targets", synapses, "N_post")
            self.variables.add_reference("_source_dt", synapses.source, "dt")
        else:
            self.variables.add_reference("_n_sources", synapses, "N_post")
            self.variables.add_reference("_n_targets", synapses, "N_pre")
            self.variables.add_reference("_source_dt", synapses.target, "dt")
        if delay is None:  # variable delays
            if getattr(synapses, "N", None) is not None:
                n_synapses = synapses.N
            else:
                n_synapses = 0
            self.variables.add_dynamic_array(
                "delay", dimensions=second.dim, size=n_synapses, constant=True
            )
            # Register the object with the `SynapticIndex` object so it gets
            # automatically resized
            synapses.register_variable(self.variables["delay"])
        else:
            if not isinstance(delay, Quantity):
                raise TypeError(
                    f"Cannot set the delay for pathway '{objname}': "
                    f"expected a quantity, got {type(delay)} instead."
                )
            if delay.size != 1:
                raise TypeError(
                    f"Cannot set the delay for pathway '{objname}': "
                    "expected a scalar quantity, got a "
                    f"quantity with shape {delay.shape!s} instead."
                )
            fail_for_dimension_mismatch(
                delay,
                second,
                "Delay has to be specified in units of seconds but got {value}",
                value=delay,
            )
            # We use a "dynamic" array of constant size here because it makes
            # the generated code easier, we don't need to deal with a different
            # type for scalar and variable delays
            self.variables.add_dynamic_array(
                "delay", dimensions=second.dim, size=1, constant=True, scalar=True
            )
            # Since this array does not grow with the number of synapses, we
            # have to resize it ourselves
            self.variables["delay"].resize(1)
            self.variables["delay"].set_value(delay)

        self._delays = self.variables["delay"]

        # Re-extract the last part of the name from the full name
        self.objname = self.name[len(synapses.name) + 1 :]

        #: The `CodeObject` initalising the `SpikeQueue` at the begin of a run
        self._initialise_queue_codeobj = None

        self.namespace = synapses.namespace

        # Allow the use of string expressions referring to synaptic (including
        # pre-/post-synaptic) variables
        # Only include non-private variables (and their indices)
        synaptic_vars = {
            varname
            for varname in list(synapses.variables)
            if not varname.startswith("_")
        }
        synaptic_idcs = {
            varname: synapses.variables.indices[varname] for varname in synaptic_vars
        }
        synaptic_vars |= {
            index_name
            for index_name in synaptic_idcs.values()
            if index_name not in ["_idx", "0"]
        }
        self.variables.add_references(synapses, synaptic_vars)
        self.variables.indices.update(synaptic_idcs)

        #: The `SpikeQueue`
        self.queue = get_device().spike_queue(self.source.start, self.source.stop)
        self.variables.add_object("_queue", self.queue)

        self._enable_group_attributes()

    def check_variable_write(self, variable):
        # Forward the check to the `Synapses` object (raises an error if no
        # synapse has been created yet)
        self.synapses.check_variable_write(variable)

    @device_override("synaptic_pathway_update_abstract_code")
    def update_abstract_code(self, run_namespace=None, level=0):
        if self.synapses.event_driven is not None:
            event_driven_eqs = self.synapses.event_driven
            try:
                event_driven_update = linear(event_driven_eqs, self.group.variables)
            except UnsupportedEquationsException:
                err = (
                    "Cannot solve the differential equations as "
                    "event-driven. Use (clock-driven) instead."
                )
                raise UnsupportedEquationsException(err)

            # TODO: Any way to do this more elegantly?
            event_driven_update = re.sub(
                r"\bdt\b", "(t - lastupdate)", event_driven_update
            )

            self.abstract_code = event_driven_update + "\n"
        else:
            self.abstract_code = ""

        self.abstract_code += self.code + "\n"
        if self.synapses.event_driven is not None:
            self.abstract_code += "lastupdate = t\n"

    @device_override("synaptic_pathway_before_run")
    def before_run(self, run_namespace):
        super().before_run(run_namespace)

    def create_code_objects(self, run_namespace):
        if self._pushspikes_codeobj is None:
            # Since this now works for general events not only spikes, we have to
            # pass the information about which variable to use to the template,
            # it can not longer simply refer to "_spikespace"
            # Strictly speaking this is only true for the standalone mode at the
            # moment, since in runtime, all the template does is to call
            # SynapticPathway.push_spike
            eventspace_name = f"_{self.event}space"
            template_kwds = {
                "eventspace_variable": self.source.variables[eventspace_name]
            }
            needed_variables = [eventspace_name]
            self._pushspikes_codeobj = create_runner_codeobj(
                self,
                "",  # no code
                "synapses_push_spikes",
                name=self.name + "_push_spikes",
                check_units=False,
                additional_variables=self.variables,
                needed_variables=needed_variables,
                template_kwds=template_kwds,
                run_namespace=run_namespace,
            )
        self.code_objects[:] = [
            weakref.proxy(self._pushspikes_codeobj),
            weakref.proxy(self.create_default_code_object(run_namespace)),
        ]

    def initialise_queue(self):
        self.eventspace = self.source.variables[self.eventspace_name].get_value()
        n_synapses = len(self.synapses)
        if n_synapses == 0 and not self.synapses._connect_called:
            raise TypeError(
                "Synapses object '%s' does not do anything, since "
                "it has not created synapses with 'connect'. "
                "Set its active attribute to False if you "
                "intend to do only do this for a subsequent"
                " run." % self.synapses.name
            )

        # Update the dt (might have changed between runs)
        self.queue.prepare(
            self._delays.get_value(),
            self.source.clock.dt_,
            self.synapse_sources.get_value(),
        )

        if (
            len({self.source.clock.dt_, self.synapses.clock.dt_, self.target.clock.dt_})
            > 1
        ):
            logger.warn(
                f"Note that the synaptic pathway '{self.name}' will run on the "
                f"clock of the group '{self.source.name}' using a dt of "
                f"{self.source.clock.dt}. Either the Synapses object "
                f"'{self.synapses.name}' or the target '{self.target.name}' "
                "(or both) are using a different dt. This might lead to "
                "unexpected results. In particular, all delays will be "
                f"rounded to multiples of {self.source.clock.dt}. If in "
                f"doubt, try to ensure that '{self.source.name}', "
                f"'{self.synapses.name}', and '{self.target.name}' use the "
                "same dt.",
                "synapses_dt_mismatch",
                once=True,
            )

    def _full_state(self):
        state = super()._full_state()
        if self.queue is not None:
            state["_spikequeue"] = self.queue._full_state()
        else:
            state["_spikequeue"] = None
        return state

    def _restore_from_full_state(self, state):
        # We have to handle the SpikeQueue separately from the other state
        # variables, so remove it from the state dictionary so that it does not
        # get treated as a state variable by the standard mechanism in
        # `VariableOwner`
        queue_state = state.pop("_spikequeue")
        super()._restore_from_full_state(state)
        if self.queue is None:
            self.queue = get_device().spike_queue(self.source.start, self.source.stop)
        self.queue._restore_from_full_state(queue_state)
        # Put the spike queue state back for future restore calls
        state["_spikequeue"] = queue_state

    def push_spikes(self):
        # Push new events (e.g. spikes) into the queue
        events = self.eventspace[: self.eventspace[len(self.eventspace) - 1]]

        if len(events):
            self.queue.push(events)


def slice_to_test(x):
    """
    Returns a testing function corresponding to whether an index is in slice x.
    x can also be an int.
    """
    try:
        x = int(x)
        return lambda y: (y == x)
    except TypeError:
        pass

    if isinstance(x, slice):
        if isinstance(x, slice) and x == slice(None):
            # No need for testing
            return lambda y: np.repeat(True, len(y))
        start, stop, step = x.start, x.stop, x.step

        if start is None:
            # No need to test for >= start
            if step is None:
                # Only have a stop value
                return lambda y: (y < stop)
            else:
                # Stop and step
                return lambda y: (y < stop) & ((y % step) == 0)
        else:
            # We need to test for >= start
            if step is None:
                if stop is None:
                    # Only a start value
                    return lambda y: (y >= start)
                else:
                    # Start and stop
                    return lambda y: (y >= start) & (y < stop)
            else:
                if stop is None:
                    # Start and step value
                    return lambda y: (y >= start) & ((y - start) % step == 0)
                else:
                    # Start, step and stop
                    return (
                        lambda y: (y >= start) & ((y - start) % step == 0) & (y < stop)
                    )
    else:
        raise TypeError(f"Expected int or slice, got {type(x)} instead")


def find_synapses(index, synaptic_neuron):
    try:
        index = index.item()
    except (TypeError, ValueError):
        pass

    if isinstance(index, (int, slice)):
        test = slice_to_test(index)
        found = test(synaptic_neuron)
        synapses = np.flatnonzero(found)
    else:
        synapses = []
        for neuron in index:
            targets = np.flatnonzero(synaptic_neuron == neuron)
            synapses.extend(targets)
        synapses = np.array(synapses, dtype=np.int32)

    return synapses


class SynapticSubgroup:
    """
    A simple subgroup of `Synapses` that can be used for indexing.

    Parameters
    ----------
    indices : `ndarray` of int
        The synaptic indices represented by this subgroup.
    synaptic_pre : `DynamicArrayVariable`
        References to all pre-synaptic indices. Only used to throw an error
        when new synapses where added after creating this object.
    """

    def __init__(self, synapses, indices):
        self.synapses = weakproxy_with_fallback(synapses)
        self._stored_indices = indices
        self._synaptic_pre = synapses.variables["_synaptic_pre"]
        self._source_N = self._synaptic_pre.size  # total number of synapses

    def _indices(self, index_var="_idx"):
        if index_var != "_idx":
            raise AssertionError(f"Did not expect index {index_var} here.")
        if len(self._synaptic_pre.get_value()) != self._source_N:
            raise RuntimeError(
                "Synapses have been added/removed since this "
                "synaptic subgroup has been created"
            )
        return self._stored_indices

    def __len__(self):
        return len(self._stored_indices)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}, storing {len(self._stored_indices):d} "
            f"indices of {self.synapses.name}>"
        )


class SynapticIndexing:
    def __init__(self, synapses):
        self.synapses = weakref.proxy(synapses)
        self.source = weakproxy_with_fallback(self.synapses.source)
        self.target = weakproxy_with_fallback(self.synapses.target)
        self.synaptic_pre = synapses.variables["_synaptic_pre"]
        self.synaptic_post = synapses.variables["_synaptic_post"]
        if synapses.multisynaptic_index is not None:
            self.synapse_number = synapses.variables[synapses.multisynaptic_index]
        else:
            self.synapse_number = None

    def __call__(self, index=None, index_var="_idx"):
        """
        Returns synaptic indices for `index`, which can be a tuple of indices
        (including arrays and slices), a single index or a string.

        """
        if index is None or (isinstance(index, str) and index == "True"):
            index = slice(None)

        if not isinstance(index, (tuple, str)) and (
            isinstance(index, (numbers.Integral, np.ndarray, slice, Sequence))
            or hasattr(index, "_indices")
        ):
            if hasattr(index, "_indices"):
                final_indices = index._indices(index_var=index_var).astype(np.int32)
            elif isinstance(index, slice):
                start, stop, step = index.indices(len(self.synaptic_pre.get_value()))
                final_indices = np.arange(start, stop, step, dtype=np.int32)
            else:
                final_indices = np.asarray(index)
        elif isinstance(index, tuple):
            if len(index) == 2:  # two indices (pre- and postsynaptic cell)
                index = (index[0], index[1], slice(None))
            elif len(index) > 3:
                raise IndexError(f"Need 1, 2 or 3 indices, got {len(index)}.")

            i_indices, j_indices, k_indices = index
            # Convert to absolute indices (e.g. for subgroups)
            # Allow the indexing to fail, we'll later return an empty array in
            # that case
            try:
                if hasattr(
                    i_indices, "_indices"
                ):  # will return absolute indices already
                    i_indices = i_indices._indices()
                else:
                    i_indices = self.source._indices(i_indices)
                pre_synapses = find_synapses(i_indices, self.synaptic_pre.get_value())
            except IndexError:
                pre_synapses = np.array([], dtype=np.int32)
            try:
                if hasattr(j_indices, "_indices"):
                    j_indices = j_indices._indices()
                else:
                    j_indices = self.target._indices(j_indices)
                post_synapses = find_synapses(j_indices, self.synaptic_post.get_value())
            except IndexError:
                post_synapses = np.array([], dtype=np.int32)

            matching_synapses = np.intersect1d(
                pre_synapses, post_synapses, assume_unique=True
            )

            if isinstance(k_indices, slice) and k_indices == slice(None):
                final_indices = matching_synapses
            else:
                if self.synapse_number is None:
                    raise IndexError(
                        "To index by the third dimension you need "
                        "to switch on the calculation of the "
                        "'multisynaptic_index' when you create "
                        "the Synapses object."
                    )
                if isinstance(k_indices, (numbers.Integral, slice)):
                    test_k = slice_to_test(k_indices)
                else:
                    raise NotImplementedError(
                        "Indexing synapses with arrays notimplemented yet"
                    )

                # We want to access the raw arrays here, not go through the Variable
                synapse_numbers = self.synapse_number.get_value()[matching_synapses]
                final_indices = np.intersect1d(
                    matching_synapses,
                    np.flatnonzero(test_k(synapse_numbers)),
                    assume_unique=True,
                )
        else:
            raise IndexError(f"Unsupported index type {type(index)}")

        if index_var not in ("_idx", "0"):
            return index_var.get_value()[final_indices.astype(np.int32)]
        else:
            return final_indices.astype(np.int32)


class Synapses(Group):
    """
    Class representing synaptic connections.

    Creating a new `Synapses` object does by default not create any synapses,
    you have to call the `Synapses.connect` method for that.

    Parameters
    ----------

    source : `SpikeSource`
        The source of spikes, e.g. a `NeuronGroup`.
    target : `Group`, optional
        The target of the spikes, typically a `NeuronGroup`. If none is given,
        the same as `source`
    model : `str`, `Equations`, optional
        The model equations for the synapses.
    on_pre : str, dict, optional
        The code that will be executed after every pre-synaptic spike. Can be
        either a single (possibly multi-line) string, or a dictionary mapping
        pathway names to code strings. In the first case, the pathway will be
        called ``pre`` and made available as an attribute of the same name.
        In the latter case, the given names will be used as the
        pathway/attribute names. Each pathway has its own code and its own
        delays.
    pre : str, dict, optional
        Deprecated. Use ``on_pre`` instead.
    on_post : str, dict, optional
        The code that will be executed after every post-synaptic spike. Same
        conventions as for `on_pre``, the default name for the pathway is
        ``post``.
    post : str, dict, optional
        Deprecated. Use ``on_post`` instead.
    delay : `Quantity`, dict, optional
        The delay for the "pre" pathway (same for all synapses) or a dictionary
        mapping pathway names to delays. If a delay is specified in this way
        for a pathway, it is stored as a single scalar value. It can still
        be changed afterwards, but only to a single scalar value. If you want
        to have delays that vary across synapses, do not use the keyword
        argument, but instead set the delays via the attribute of the pathway,
        e.g. ``S.pre.delay = ...`` (or ``S.delay = ...`` as an abbreviation),
        ``S.post.delay = ...``, etc.
    on_event : str or dict, optional
        Define the events which trigger the pre and post pathways. By default,
        both pathways are triggered by the ``'spike'`` event, i.e. the event
        that is triggered by the ``threshold`` condition in the connected
        groups.
    multisynaptic_index : str, optional
        The name of a variable (which will be automatically created) that stores
        the "synapse number". This number enumerates all synapses between the
        same source and target so that they can be distinguished. For models
        where each source-target pair has only a single connection, this number
        only wastes memory (it would always default to 0), it is therefore not
        stored by default. Defaults to ``None`` (no variable).
    namespace : dict, optional
        A dictionary mapping identifier names to objects. If not given, the
        namespace will be filled in at the time of the call of `Network.run`,
        with either the values from the ``namespace`` argument of the
        `Network.run` method or from the local context, if no such argument is
        given.
    dtype : `dtype`, dict, optional
        The `numpy.dtype` that will be used to store the values, or a
        dictionary specifying the type for variable names. If a value is not
        provided for a variable (or no value is provided at all), the preference
        setting `core.default_float_dtype` is used.
    codeobj_class : class, optional
        The `CodeObject` class to use to run code.
    dt : `Quantity`, optional
        The time step to be used for the update of the state variables.
        Cannot be combined with the `clock` argument.
    clock : `Clock`, optional
        The update clock to be used. If neither a clock, nor the `dt` argument
        is specified, the `defaultclock` will be used.
    order : int, optional
        The priority of of this group for operations occurring at the same time
        step and in the same scheduling slot. Defaults to 0.
    method : str, `StateUpdateMethod`, optional
        The numerical integration method to use. If none is given, an
        appropriate one is automatically determined.
    name : str, optional
        The name for this object. If none is given, a unique name of the form
        ``synapses``, ``synapses_1``, etc. will be automatically chosen.
    """

    add_to_magic_network = True

    def __init__(
        self,
        source,
        target=None,
        model=None,
        on_pre=None,
        pre=None,
        on_post=None,
        post=None,
        connect=None,
        delay=None,
        on_event="spike",
        multisynaptic_index=None,
        namespace=None,
        dtype=None,
        codeobj_class=None,
        dt=None,
        clock=None,
        order=0,
        method=("exact", "euler", "heun"),
        method_options=None,
        name="synapses*",
    ):
        if connect is not None:
            raise TypeError(
                "The connect keyword argument is no longer "
                "supported, call the connect method instead."
            )

        if pre is not None:
            if on_pre is not None:
                raise TypeError(
                    "Cannot specify both 'pre' and 'on_pre'. The "
                    "'pre' keyword is deprecated, use the 'on_pre' "
                    "keyword instead."
                )
            logger.warn(
                "The 'pre' keyword is deprecated, use 'on_pre' instead.",
                "deprecated_pre",
                once=True,
            )
            on_pre = pre

        if post is not None:
            if on_post is not None:
                raise TypeError(
                    "Cannot specify both 'post' and 'on_post'. The "
                    "'post' keyword is deprecated, use the "
                    "'on_post' keyword instead."
                )
            logger.warn(
                "The 'post' keyword is deprecated, use 'on_post' instead.",
                "deprecated_post",
                once=True,
            )
            on_post = post

        Group.__init__(
            self,
            dt=dt,
            clock=clock,
            when="start",
            order=order,
            namespace=namespace,
            name=name,
        )

        if dtype is None:
            dtype = {}
        if isinstance(dtype, MutableMapping):
            dtype["lastupdate"] = self._clock.variables["t"].dtype

        #: remember whether connect was called to raise an error if an
        #: assignment to a synaptic variable is attempted without a preceding
        #: connect.
        self._connect_called = False
        self.codeobj_class = codeobj_class

        self.source = source
        self.add_dependency(source)
        if target is None:
            self.target = self.source
        else:
            self.target = target
            self.add_dependency(target)

        ##### Prepare and validate equations
        if model is None:
            model = ""

        if isinstance(model, str):
            model = Equations(model)
        if not isinstance(model, Equations):
            raise TypeError(
                "model has to be a string or an Equations "
                f"object, is '{type(model)}' instead."
            )

        # Check flags
        model.check_flags(
            {
                DIFFERENTIAL_EQUATION: ["event-driven", "clock-driven"],
                SUBEXPRESSION: ["summed", "shared", "constant over dt"],
                PARAMETER: ["constant", "shared", "linked"],
            },
            incompatible_flags=[
                ("event-driven", "clock-driven"),
                # 'summed' cannot be combined with
                # any other flag
                ("summed", "shared", "constant over dt"),
            ],
        )

        for name in ["i", "j", "delay"]:
            if name in model.names:
                raise SyntaxError(
                    f"'{name}' is a reserved name that cannot be "
                    "used as a variable name."
                )

        # Add the "multisynaptic index", if desired
        self.multisynaptic_index = multisynaptic_index
        if multisynaptic_index is not None:
            if not isinstance(multisynaptic_index, str):
                raise TypeError("multisynaptic_index argument has to be a string")
            model = model + Equations(f"{multisynaptic_index} : integer")

        # Separate subexpressions depending whether they are considered to be
        # constant over a time step or not
        model, constant_over_dt = extract_constant_subexpressions(model)
        # Separate the equations into event-driven equations,
        # continuously updated equations and summed variable updates
        event_driven = []
        continuous = []
        summed_updates = []
        for single_equation in model.values():
            if "event-driven" in single_equation.flags:
                event_driven.append(single_equation)
            elif "summed" in single_equation.flags:
                summed_updates.append(single_equation)
            else:
                if (
                    single_equation.type == DIFFERENTIAL_EQUATION
                    and "clock-driven" not in single_equation.flags
                ):
                    logger.info(
                        "The synaptic equation for the variable "
                        f"{single_equation.varname} does not specify whether it "
                        "should be integrated at every timestep ('clock-driven') "
                        "or only at spiking events ('event-driven'). It will be "
                        "integrated at every timestep which can slow down your "
                        "simulation unnecessarily if you only need the values of "
                        "this variable whenever a spike occurs. Specify the equation "
                        "as clock-driven explicitly to avoid this warning.",
                        "clock_driven",
                        once=True,
                    )
                continuous.append(single_equation)
                if single_equation.type != DIFFERENTIAL_EQUATION:
                    # General subexpressions (not summed variables) or
                    # parameters, might be referred from event-driven equations
                    # as well.
                    # Note that the code generation step will ignore them if
                    # nothing refers to them, so we don't have to filter here.
                    event_driven.append(single_equation)
        # Get the dependencies of all equations
        dependencies = model.dependencies
        # Check whether there are dependencies between summed
        # variables/clocked-driven equations and event-driven variables
        for eq_name, deps in dependencies.items():
            eq = model[eq_name]
            if not (eq.type == DIFFERENTIAL_EQUATION or "summed" in eq.flags):
                continue
            if eq in continuous:
                Synapses.verify_dependencies(
                    eq, "clock-driven", deps, event_driven, "event-driven"
                )
            elif "summed" in eq.flags:
                Synapses.verify_dependencies(
                    eq, "summed", deps, event_driven, "event-driven"
                )
            elif eq in event_driven:
                Synapses.verify_dependencies(
                    eq, "event-driven", deps, continuous, "clock-driven"
                )

        if any(eq.type == DIFFERENTIAL_EQUATION for eq in event_driven):
            self.event_driven = Equations(event_driven)
            # Add the lastupdate variable, needed for event-driven updates
            model += Equations("lastupdate : second")
        else:
            self.event_driven = None

        self._linked_variables = set()

        self._create_variables(model, user_dtype=dtype)
        self.equations = Equations(continuous)

        #: Set of `Variable` objects that should be resized when the
        #: number of synapses changes
        self._registered_variables = set()

        for varname, var in self.variables.items():
            if (
                isinstance(var, DynamicArrayVariable)
                and self.variables.indices[varname] == "_idx"
            ):
                # Register the array with the `SynapticItemMapping` object so
                # it gets automatically resized
                self.register_variable(var)

        # Support 2d indexing
        self._indices = SynapticIndexing(self)

        if delay is None:
            delay = {}

        if isinstance(delay, Quantity):
            delay = {"pre": delay}
        elif not isinstance(delay, Mapping):
            raise TypeError(
                "Delay argument has to be a quantity or a "
                f"dictionary, is type {type(delay)} instead."
            )

        #: List of names of all updaters, e.g. ['pre', 'post']
        self._synaptic_updaters = []
        #: List of all `SynapticPathway` objects
        self._pathways = []

        if isinstance(on_event, str):
            self.default_event = on_event
            events_dict = {}
        else:
            self.default_event = "spike"
            events_dict = dict(on_event)

        #: "Events" for all the pathways
        self.events = events_dict
        for prepost, argument in zip(("pre", "post"), (on_pre, on_post)):
            if not argument:
                continue
            if isinstance(argument, str):
                pathway_delay = delay.get(prepost, None)
                self._add_updater(
                    argument,
                    prepost,
                    delay=pathway_delay,
                    event=self.events.get(prepost, self.default_event),
                )
            elif isinstance(argument, Mapping):
                for key, value in argument.items():
                    if not isinstance(key, str):
                        err_msg = (
                            f"Keys for the 'on_{prepost}' argument"
                            "have to be strings, got "
                            f"{type(key)} instead."
                        )
                        raise TypeError(err_msg)
                    pathway_delay = delay.get(key, None)
                    self._add_updater(
                        value,
                        prepost,
                        objname=key,
                        delay=pathway_delay,
                        event=self.events.get(key, self.default_event),
                    )

        # Check whether any delays were specified for pathways that don't exist
        for pathway in delay:
            if pathway not in self._synaptic_updaters:
                raise ValueError(
                    f"Cannot set the delay for pathway '{pathway}': unknown pathway."
                )

        #: Performs numerical integration step
        self.state_updater = None

        # We only need a state update if we have differential equations
        if len(self.equations.diff_eq_names):
            self.state_updater = StateUpdater(
                self,
                method,
                method_options=method_options,
                clock=self.clock,
                order=order,
            )
            self.contained_objects.append(self.state_updater)

        #: Update the "constant over a time step" subexpressions
        self.subexpression_updater = None
        if len(constant_over_dt) > 0:
            self.subexpression_updater = SubexpressionUpdater(self, constant_over_dt)
            self.contained_objects.append(self.subexpression_updater)

        #: "Summed variable" mechanism -- sum over all synapses of a
        #: pre-/postsynaptic target
        self.summed_updaters = {}
        # We want to raise an error if the same variable is updated twice
        # using this mechanism. This could happen if the Synapses object
        # connected a NeuronGroup to itself since then all variables are
        # accessible as var_pre and var_post.
        summed_targets = set()
        for single_equation in summed_updates:
            varname = single_equation.varname
            if not (varname.endswith("_pre") or varname.endswith("_post")):
                raise ValueError(
                    f"The summed variable '{varname}' does not end "
                    "in '_pre' or '_post'."
                )
            if varname not in self.variables:
                raise ValueError(
                    f"The summed variable '{varname}' does not refer "
                    "to any known variable in the "
                    "target group."
                )
            if varname.endswith("_pre"):
                summed_target = self.source
                summed_target_size_name = "N_pre"
                orig_varname = varname[:-4]
                summed_var_index = "_synaptic_pre"
            else:
                summed_target = self.target
                summed_target_size_name = "N_post"
                orig_varname = varname[:-5]
                summed_var_index = "_synaptic_post"

            target_eq = getattr(summed_target, "equations", {}).get(orig_varname, None)
            if target_eq is None or target_eq.type != PARAMETER:
                raise ValueError(
                    f"The summed variable '{varname}' needs a "
                    f"corresponding parameter '{orig_varname}' in the "
                    "target group."
                )

            fail_for_dimension_mismatch(
                self.variables["_summed_" + varname].dim,
                self.variables[varname].dim,
                "Summed variables need to have "
                "the same units in Synapses "
                "and the target group",
            )
            if self.variables[varname] in summed_targets:
                raise ValueError(
                    f"The target variable '{orig_varname}' is already "
                    "updated by another summed variable"
                )
            summed_targets.add(self.variables[varname])
            updater = SummedVariableUpdater(
                single_equation.expr,
                varname,
                self,
                summed_target,
                summed_target_size_name,
                summed_var_index,
            )
            self.summed_updaters[varname] = updater
            self.contained_objects.append(updater)

        # Activate name attribute access
        self._enable_group_attributes()

    @staticmethod
    def verify_dependencies(
        eq, eq_type, deps, should_not_depend_on, should_not_depend_on_name
    ):
        """
        Helper function to verify that event-driven equations do not depend
        on clock-driven equations and the other way round.

        Parameters
        ----------
        eq : `SingleEquation`
            The equation to verify
        eq_type : str
            The type of the equation (for the error message)
        deps : list
            A list of dependencies
        should_not_depend_on : list
            A list of equations to verify against the dependencies
        should_not_depend_on_name : str
            The name of the list of equations (for the error message)

        Raises
        ------
        `EquationError`
            If the given equation depends on something in the other set of
            equations.
        """
        for dep in deps:
            if dep.equation in should_not_depend_on and (
                dep.equation.type == DIFFERENTIAL_EQUATION
                or "summed" in dep.equation.flags
            ):
                via_str = ""
                if dep.via:
                    via_str = " (via " + ", ".join(f"'{v}'" for v in dep.via) + ")"
                raise EquationError(
                    f"The {eq_type} '{eq.varname}' should "
                    "not depend on the "
                    f"{should_not_depend_on_name} variable "
                    f"'{dep.equation.varname}'{via_str}."
                )

    N_outgoing_pre = property(
        fget=lambda self: self.variables["N_outgoing"].get_value(),
        doc=(
            "The number of outgoing synapses for each neuron in the pre-synaptic group."
        ),
    )
    N_incoming_post = property(
        fget=lambda self: self.variables["N_incoming"].get_value(),
        doc=(
            "The number of incoming synapses for each neuron in the "
            "post-synaptic group."
        ),
    )

    def __getitem__(self, item):
        indices = self.indices[item]
        return SynapticSubgroup(self, indices)

    def _set_delay(self, delay, with_unit):
        if "pre" not in self._synaptic_updaters:
            raise AttributeError(
                "Synapses do not have a 'pre' pathway, "
                "do not know what 'delay' refers to."
            )
        # Note that we cannot simply say: "self.pre.delay = delay" because this
        # would not correctly deal with references to external constants
        var = self.pre.variables["delay"]
        if with_unit:
            reference = var.get_addressable_value_with_unit("delay", self.pre)
        else:
            reference = var.get_addressable_value("delay", self.pre)
        reference.set_item("True", delay, level=2)

    def _get_delay(self, with_unit):
        if "pre" not in self._synaptic_updaters:
            raise AttributeError(
                "Synapses do not have a 'pre' pathway, "
                "do not know what 'delay' refers to."
            )
        var = self.pre.variables["delay"]
        if with_unit:
            return var.get_addressable_value_with_unit("delay", self.pre)
        else:
            return var.get_addressable_value("delay", self.pre)

    delay = property(
        functools.partial(_get_delay, with_unit=True),
        functools.partial(_set_delay, with_unit=True),
        doc="The presynaptic delay (if a pre-synaptic pathway exists).",
    )
    delay_ = property(
        functools.partial(_get_delay, with_unit=False),
        functools.partial(_set_delay, with_unit=False),
        doc=(
            "The presynaptic delay without unit information (if a"
            "pre-synaptic pathway exists)."
        ),
    )

    def _add_updater(self, code, prepost, objname=None, delay=None, event="spike"):
        """
        Add a new target updater. Users should call `add_pre` or `add_post`
        instead.

        Parameters
        ----------
        code : str
            The abstract code that should be executed on pre-/postsynaptic
            spikes.
        prepost : {'pre', 'post'}
            Whether the code is triggered by presynaptic or postsynaptic spikes
        objname : str, optional
            A name for the object, see `SynapticPathway` for more details.
        delay : `Quantity`, optional
            A scalar delay (same delay for all synapses) for this pathway. If
            not given, delays are expected to vary between synapses.

        Returns
        -------
        objname : str
            The final name for the object. Equals `objname` if it was explicitly
            given (and did not end in a wildcard character).

        """
        if prepost == "pre":
            spike_group, group_name = self.source, "Source"
        elif prepost == "post":
            spike_group, group_name = self.target, "Target"
        else:
            raise AssertionError(
                f"'prepost' argument has to be 'pre' or 'post', is '{prepost}'."
            )
        if event not in spike_group.events:
            if event == "spike":
                threshold_text = " Did you forget to set a 'threshold'?"
            else:
                threshold_text = ""
            raise ValueError(
                f"{group_name} group '{spike_group.name}' does not define "
                f"an event '{event}'.{threshold_text}"
            )

        if not isinstance(spike_group, SpikeSource) or not hasattr(
            spike_group, "clock"
        ):
            raise TypeError(
                f"'{group_name}' has to be a SpikeSource with spikes and"
                f" clock attribute. Is type {type(spike_group)!r} instead."
            )

        updater = SynapticPathway(
            self, code, prepost, objname, delay=delay, event=event
        )
        objname = updater.objname
        if hasattr(self, objname):
            raise ValueError(
                f"Cannot add updater with name '{objname}', synapses "
                "object already has an attribute with this "
                "name."
            )

        setattr(self, objname, updater)
        self._synaptic_updaters.append(objname)
        self._pathways.append(updater)
        self.contained_objects.append(updater)
        return objname

    def _create_variables(self, equations, user_dtype=None):
        """
        Create the variables dictionary for this `Synapses`, containing
        entries for the equation variables and some standard entries.
        """
        self.variables = Variables(self)

        # Standard variables always present
        self.variables.add_dynamic_array(
            "_synaptic_pre", size=0, dtype=np.int32, constant=True, read_only=True
        )
        self.variables.add_dynamic_array(
            "_synaptic_post", size=0, dtype=np.int32, constant=True, read_only=True
        )
        self.variables.create_clock_variables(self._clock)
        if "_offset" in self.target.variables:
            self.variables.add_reference("_target_offset", self.target, "_offset")
        else:
            self.variables.add_constant("_target_offset", value=0)
        if "_offset" in self.source.variables:
            self.variables.add_reference("_source_offset", self.source, "_offset")
        else:
            self.variables.add_constant("_source_offset", value=0)
        # To cope with connections to/from other synapses, N_incoming/N_outgoing
        # will be resized when synapses are created
        self.variables.add_dynamic_array(
            "N_incoming",
            size=0,
            dtype=np.int32,
            constant=True,
            read_only=True,
            index="_postsynaptic_idx",
        )
        self.variables.add_dynamic_array(
            "N_outgoing",
            size=0,
            dtype=np.int32,
            constant=True,
            read_only=True,
            index="_presynaptic_idx",
        )

        # We have to make a distinction here between the indices
        # and the arrays (even though they refer to the same object)
        # the synaptic propagation template would otherwise overwrite
        # synaptic_post in its namespace with the value of the
        # postsynaptic index, leading to errors for the next
        # propagation.
        self.variables.add_reference("_presynaptic_idx", self, "_synaptic_pre")
        self.variables.add_reference("_postsynaptic_idx", self, "_synaptic_post")

        # Except for subgroups (which potentially add an offset), the "i" and
        # "j" variables are simply equivalent to `_synaptic_pre` and
        # `_synaptic_post`
        if getattr(self.source, "start", 0) == 0:
            self.variables.add_reference("i", self, "_synaptic_pre")
        else:
            self.variables.add_reference(
                "_source_i", self.source.source, "i", index="_presynaptic_idx"
            )
            self.variables.add_reference("_source_offset", self.source, "_offset")
            self.variables.add_subexpression(
                "i",
                dtype=self.source.source.variables["i"].dtype,
                expr="_source_i - _source_offset",
                index="_presynaptic_idx",
            )
        if getattr(self.target, "start", 0) == 0:
            self.variables.add_reference("j", self, "_synaptic_post")
        else:
            self.variables.add_reference(
                "_target_j", self.target.source, "i", index="_postsynaptic_idx"
            )
            self.variables.add_reference("_target_offset", self.target, "_offset")
            self.variables.add_subexpression(
                "j",
                dtype=self.target.source.variables["i"].dtype,
                expr="_target_j - _target_offset",
                index="_postsynaptic_idx",
            )

        # Add the standard variables
        self.variables.add_array(
            "N", dtype=np.int32, size=1, scalar=True, constant=True, read_only=True
        )

        for eq in equations.values():
            dtype = get_dtype(eq, user_dtype)
            if eq.type in (DIFFERENTIAL_EQUATION, PARAMETER):
                check_identifier_pre_post(eq.varname)
                constant = "constant" in eq.flags
                shared = "shared" in eq.flags
                linked = "linked" in eq.flags
                if linked:
                    self._linked_variables.add(eq.varname)
                elif shared:
                    self.variables.add_array(
                        eq.varname,
                        size=1,
                        dimensions=eq.dim,
                        dtype=dtype,
                        constant=constant,
                        scalar=True,
                        index="0",
                    )
                else:
                    self.variables.add_dynamic_array(
                        eq.varname,
                        size=0,
                        dimensions=eq.dim,
                        dtype=dtype,
                        constant=constant,
                    )
            elif eq.type == SUBEXPRESSION:
                if "summed" in eq.flags:
                    # Give a special name to the subexpression for summed
                    # variables to avoid confusion with the pre/postsynaptic
                    # target variable
                    varname = "_summed_" + eq.varname
                else:
                    check_identifier_pre_post(eq.varname)
                    varname = eq.varname
                self.variables.add_subexpression(
                    varname,
                    dimensions=eq.dim,
                    expr=str(eq.expr),
                    scalar="shared" in eq.flags,
                    dtype=dtype,
                )
            else:
                raise AssertionError(f"Unknown type of equation: {eq.eq_type}")

        # Stochastic variables
        for xi in equations.stochastic_variables:
            self.variables.add_auxiliary_variable(xi, dimensions=(second**-0.5).dim)

        # Add all the pre and post variables with _pre and _post suffixes
        for name in getattr(self.source, "variables", {}):
            # Raise an error if a variable name is also used for a synaptic
            # variable (we ignore 'lastupdate' to allow connections from another
            # Synapses object)
            if (
                name in equations.names
                and name != "lastupdate"
                and "summed" not in equations[name].flags
            ):
                error_msg = (
                    f"The pre-synaptic variable {name} has the same "
                    "name as a synaptic variable, rename the synaptic "
                    "variable."
                )
                if name + "_syn" not in self.variables:
                    error_msg += f"(for example to '{name}_syn') "
                error_msg += "to avoid confusion"
                raise ValueError(error_msg)
            if name.startswith("_"):
                continue  # Do not add internal variables
            var = self.source.variables[name]
            index = "0" if var.scalar else "_presynaptic_idx"
            try:
                self.variables.add_reference(
                    name + "_pre", self.source, name, index=index
                )
            except TypeError:
                logger.diagnostic(
                    f"Cannot include a reference to '{name}' in "
                    f"'{self.name}', '{name}' uses a non-standard "
                    "indexing in the pre-synaptic group "
                    f"'{self.source.name}'."
                )
        for name in getattr(self.target, "variables", {}):
            # Raise an error if a variable name is also used for a synaptic
            # variable (we ignore 'lastupdate' to allow connections to another
            # Synapses object)
            if (
                name in equations.names
                and name != "lastupdate"
                and "summed" not in equations[name].flags
            ):
                error_msg = (
                    f"The post-synaptic variable '{name}' has the same "
                    "name as a synaptic variable, rename the synaptic "
                    "variable."
                )
                if name + "_syn" not in self.variables:
                    error_msg += f"(for example to '{name}_syn') "
                error_msg += "to avoid confusion"
                raise ValueError(error_msg)
            if name.startswith("_"):
                continue  # Do not add internal variables
            var = self.target.variables[name]
            index = "0" if var.scalar else "_postsynaptic_idx"
            try:
                self.variables.add_reference(
                    name + "_post", self.target, name, index=index
                )
                # Also add all the post variables without a suffix, but only if
                # it does not have a post or pre suffix in the target group
                # (which could happen when connecting to synapses)
                if not name.endswith("_post") or name.endswith("_pre"):
                    self.variables.add_reference(name, self.target, name, index=index)
            except TypeError:
                logger.diagnostic(
                    f"Cannot include a reference to '{name}' in "
                    f"'{self.name}', '{name}' uses a non-standard "
                    "indexing in the post-synaptic group "
                    f"'{self.target.name}'."
                )

        # Check scalar subexpressions
        for eq in equations.values():
            if eq.type == SUBEXPRESSION and "shared" in eq.flags:
                var = self.variables[eq.varname]
                for identifier in var.identifiers:
                    if identifier in self.variables:
                        if not self.variables[identifier].scalar:
                            raise SyntaxError(
                                f"Shared subexpression '{eq.varname}' "
                                "refers to non-shared variable "
                                f"'{identifier}'."
                            )

    def before_run(self, run_namespace):
        self.equations.check_units(self, run_namespace=run_namespace)
        # Check that subexpressions that refer to stateful functions are labeled
        # as "constant over dt"
        check_subexpressions(self, self.equations, run_namespace)
        super().before_run(run_namespace=run_namespace)

    @device_override("synapses_connect")
    def connect(
        self,
        condition=None,
        i=None,
        j=None,
        p=1.0,
        n=1,
        skip_if_invalid=False,
        namespace=None,
        level=0,
    ):
        """
        Add synapses.

        See :doc:`/user/synapses` for details.

        Parameters
        ----------
        condition : str, bool, optional
            A boolean or string expression that evaluates to a boolean.
            The expression can depend on indices ``i`` and ``j`` and on
            pre- and post-synaptic variables. Can be combined with
            arguments ``n``, and ``p`` but not ``i`` or ``j``.
        i : int, ndarray of int, str, optional
            The presynaptic neuron indices  It can be an index or array of
            indices if combined with the ``j`` argument, or it can be a string
            generator expression.
        j : int, ndarray of int, str, optional
            The postsynaptic neuron indices. It can be an index or array of
            indices if combined with the ``i`` argument, or it can be a string
            generator expression.
        p : float, str, optional
            The probability to create ``n`` synapses wherever the ``condition``
            evaluates to true. Cannot be used with generator syntax for ``j``.
        n : int, str, optional
            The number of synapses to create per pre/post connection pair.
            Defaults to 1.
        skip_if_invalid : bool, optional
            If set to True, rather than raising an error if you try to
            create an invalid/out of range pair (i, j) it will just
            quietly skip those synapses.
        namespace : dict-like, optional
            A namespace that will be used in addition to the group-specific
            namespaces (if defined). If not specified, the locals
            and globals around the run function will be used.
        level : int, optional
            How deep to go up the stack frame to look for the locals/global
            (see ``namespace`` argument).

        Examples
        --------
        >>> from brian2 import *
        >>> import numpy as np
        >>> G = NeuronGroup(10, 'dv/dt = -v / tau : 1', threshold='v>1', reset='v=0')
        >>> S = Synapses(G, G, 'w:1', on_pre='v+=w')
        >>> S.connect(condition='i != j') # all-to-all but no self-connections
        >>> S.connect(i=0, j=0) # connect neuron 0 to itself
        >>> S.connect(i=np.array([1, 2]), j=np.array([2, 1])) # connect 1->2 and 2->1
        >>> S.connect() # connect all-to-all
        >>> S.connect(condition='i != j', p=0.1)  # Connect neurons with 10% probability, exclude self-connections
        >>> S.connect(j='i', n=2)  # Connect all neurons to themselves with 2 synapses
        >>> S.connect(j='k for k in range(i+1)') # Connect neuron i to all j with 0<=j<=i
        >>> S.connect(j='i+(-1)**k for k in range(2) if i>0 and i<N_pre-1') # connect neuron i to its neighbours if it has both neighbours
        >>> S.connect(j='k for k in sample(N_post, p=i*1.0/(N_pre-1))') # neuron i connects to j with probability i/(N-1)
        >>> S.connect(j='k for k in sample(N_post, size=i//2)') # Each neuron connects to i//2 other neurons (chosen randomly)
        """
        # check types
        self._verify_connect_argument_types(condition, i, j, n, p)

        self._connect_called = True

        # Get namespace information
        if namespace is None:
            namespace = get_local_namespace(level=level + 2)

        try:  # wrap everything to catch IndexError
            # which connection case are we in?
            # 1: Connection condition
            if condition is None and i is None and j is None:
                condition = True
            if condition is not None:
                if i is not None or j is not None:
                    raise ValueError("Cannot combine condition with i or j arguments")
                if condition is False or condition == "False":
                    # Nothing to do
                    return
                j = self._condition_to_generator_expression(condition, p, namespace)
                self._add_synapses_generator(
                    j,
                    n,
                    skip_if_invalid=skip_if_invalid,
                    namespace=namespace,
                    level=level + 2,
                    over_presynaptic=True,
                )
            # 2: connection indices
            elif (i is not None and j is not None) and not (
                isinstance(i, str) or isinstance(j, str)
            ):
                if skip_if_invalid:
                    raise ValueError("Can only use skip_if_invalid with string syntax")
                i, j, n = self._verify_connect_array_arguments(i, j, n)
                self._add_synapses_from_arrays(i, j, n, p, namespace=namespace)
            # 3: Generator expression over post-synaptic cells (i='...')
            elif isinstance(i, str):
                i = self._finalize_generator_expression(i, j, p, "i", "j")
                self._add_synapses_generator(
                    i,
                    n,
                    skip_if_invalid=skip_if_invalid,
                    namespace=namespace,
                    level=level + 2,
                    over_presynaptic=False,
                )
            # 4: Generator expression over pre-synaptic cells (i='...')
            elif isinstance(j, str):
                j = self._finalize_generator_expression(j, i, p, "j", "i")
                self._add_synapses_generator(
                    j,
                    n,
                    skip_if_invalid=skip_if_invalid,
                    namespace=namespace,
                    level=level + 2,
                    over_presynaptic=True,
                )
            else:
                raise ValueError(
                    "Must specify at least one of condition, i or j arguments"
                )
        except IndexError as e:
            raise IndexError(
                "Tried to create synapse indices outside valid "
                "range. Original error message: " + str(e)
            )

    # Helper functions for Synapses.connect ↑
    def _verify_connect_array_arguments(self, i, j, n):
        if hasattr(i, "_indices"):
            i = i._indices()
        i = np.asarray(i)
        if not np.issubdtype(i.dtype, np.signedinteger):
            raise TypeError(
                "Presynaptic indices have to be given as "
                f"integers, are type {i.dtype} "
                "instead."
            )
        if hasattr(j, "_indices"):
            j = j._indices()
        j = np.asarray(j)
        if not np.issubdtype(j.dtype, np.signedinteger):
            raise TypeError(
                "Presynaptic indices can only be combined "
                "with postsynaptic integer indices))"
            )
        if isinstance(n, str):
            raise TypeError(
                "Indices cannot be combined with a string"
                "expression for n. Either use an "
                "array/scalar for n, or a string "
                "expression for the connections"
            )
        i, j, n = np.broadcast_arrays(i, j, n)
        if i.ndim > 1:
            raise ValueError("Can only use 1-dimensional indices")
        return i, j, n

    def _condition_to_generator_expression(self, condition, p, namespace):
        if condition is True:
            condition = "True"
        # Check that the condition is a boolean expresion
        identifiers = get_identifiers(condition)
        variables = self.resolve_all(identifiers, namespace)
        if not is_boolean_expression(condition, variables):
            raise TypeError(f"Condition '{condition}' is not a boolean condition")
        # Check the units (mostly to check for unit consistency within the condition)
        dims = parse_expression_dimensions(condition, variables)
        if dims is not DIMENSIONLESS:
            # We should not get here normally
            raise TypeError(f"Condition '{condition}' is not a boolean condition")
        condition = word_substitute(condition, {"j": "_k"})
        if not isinstance(p, str) and p == 1:
            j = f"_k for _k in range(N_post) if {condition}"
        else:
            j = None
            if isinstance(p, str):
                identifiers = get_identifiers(p)
                variables = self.resolve_all(identifiers, namespace)
                dim = parse_expression_dimensions(p, variables)
                if dim is not DIMENSIONLESS:
                    raise DimensionMismatchError(
                        "Expression for p should be dimensionless."
                    )
                p_dep = self._expression_index_dependence(p, namespace=namespace)
                if "_postsynaptic_idx" in p_dep or "_iterator_idx" in p_dep:
                    j = f"_k for _k in range(N_post) if ({condition}) and rand()<{p}"
            if j is None:
                j = f"_k for _k in sample(N_post, p={p}) if {condition}"
        return j

    def _verify_connect_argument_types(self, condition, i, j, n, p):
        if condition is not None and not isinstance(condition, (bool, str)):
            raise TypeError(
                "condition argument must be bool or string. If you "
                "want to connect based on indices, use "
                "connect(i=..., j=...)."
            )
        if i is not None and not (
            isinstance(i, (numbers.Integral, np.ndarray, Sequence))
            or hasattr(i, "_indices")
        ):
            raise TypeError("i argument must be int, array or string")
        if j is not None and not (
            isinstance(j, (numbers.Integral, np.ndarray, Sequence))
            or hasattr(j, "_indices")
        ):
            raise TypeError("j argument must be int, array or string")
        # TODO: eliminate these restrictions
        if not isinstance(p, (int, float, str)):
            raise TypeError("p must be float or string")
        if not isinstance(n, (int, str)):
            raise TypeError("n must be int or string")
        if isinstance(condition, str) and re.search(r"\bfor\b", condition):
            raise ValueError(
                "Generator expression given for condition, write "
                f"connect(j='{condition}'...) instead of "
                f"connect('{condition}'...)."
            )

    def check_variable_write(self, variable):
        """
        Checks that `Synapses.connect` has been called before setting a
        synaptic variable.

        Parameters
        ----------
        variable : `Variable`
            The variable that the user attempts to set.

        Raises
        ------
        TypeError
            If `Synapses.connect` has not been called yet.
        """
        if not self._connect_called:
            raise TypeError(
                f"Cannot write to synaptic variable '{variable.name}', you "
                "need to call connect(...) first"
            )

    def _resize(self, number):
        if not isinstance(number, (numbers.Integral, np.integer)):
            raise TypeError(f"Expected an integer number, got {type(number)} instead.")
        if number < self.N:
            raise ValueError(
                f"Cannot reduce number of synapses, {number} < {len(self)}."
            )

        for variable in self._registered_variables:
            variable.resize(number)

        self.variables["N"].set_value(number)

    def _update_synapse_numbers(self, old_num_synapses):
        source_offset = self.variables["_source_offset"].get_value()
        target_offset = self.variables["_target_offset"].get_value()
        # This resizing is only necessary if we are connecting to/from synapses
        post_with_offset = self.variables["N_post"].item() + target_offset
        pre_with_offset = self.variables["N_pre"].item() + source_offset
        self.variables["N_incoming"].resize(post_with_offset)
        self.variables["N_outgoing"].resize(pre_with_offset)
        N_outgoing = self.variables["N_outgoing"].get_value()
        N_incoming = self.variables["N_incoming"].get_value()
        synaptic_pre = self.variables["_synaptic_pre"].get_value()
        synaptic_post = self.variables["_synaptic_post"].get_value()

        # Update the number of total outgoing/incoming synapses per
        # source/target neuron
        N_outgoing[:] += np.bincount(
            synaptic_pre[old_num_synapses:], minlength=len(N_outgoing)
        )
        N_incoming[:] += np.bincount(
            synaptic_post[old_num_synapses:], minlength=len(N_incoming)
        )

        if self.multisynaptic_index is not None:
            synapse_number_var = self.variables[self.multisynaptic_index]
            synapse_number = synapse_number_var.get_value()

            # Update the "synapse number" (number of synapses for the same
            # source-target pair)
            # We wrap pairs of source/target indices into a complex number for
            # convenience
            _source_target_pairs = synaptic_pre + synaptic_post * 1j
            synapse_number[:] = calc_repeats(_source_target_pairs)

    def register_variable(self, variable):
        """
        Register a `DynamicArray` to be automatically resized when the size of
        the indices change. Called automatically when a `SynapticArrayVariable`
        specifier is created.
        """
        if not hasattr(variable, "resize"):
            raise TypeError(
                f"Variable of type {type(variable)} does not have a resize "
                "method, cannot register it with the synaptic "
                "indices."
            )
        self._registered_variables.add(variable)

    def unregister_variable(self, variable):
        """
        Unregister a `DynamicArray` from the automatic resizing mechanism.
        """
        self._registered_variables.remove(variable)

    def _get_multisynaptic_indices(self):
        template_kwds = {"multisynaptic_index": self.multisynaptic_index}
        if self.multisynaptic_index is not None:
            needed_variables = [self.multisynaptic_index]
        else:
            needed_variables = []
        return template_kwds, needed_variables

    def _add_synapses_from_arrays(self, sources, targets, n, p, namespace=None):
        template_kwds, needed_variables = self._get_multisynaptic_indices()

        variables = Variables(self)

        sources = np.atleast_1d(sources).astype(np.int32)
        targets = np.atleast_1d(targets).astype(np.int32)

        # Check whether the values in sources/targets make sense
        error_message = (
            "The given {source_or_target} indices contain "
            "values outside of the range [0, {max_value}] "
            "allowed for the {source_or_target} group "
            '"{group_name}"'
        )
        try:
            for indices, source_or_target, group in [
                (sources, "source", self.source),
                (targets, "target", self.target),
            ]:
                if np.max(indices) >= len(group) or np.min(indices) < 0:
                    raise IndexError(
                        error_message.format(
                            source_or_target=source_or_target,
                            max_value=len(group) - 1,
                            group_name=group.name,
                        )
                    )
        except NotImplementedError:
            logger.warn(
                "Cannot check whether the indices given for the connect call are valid."
                " This can happen in standalone mode when using indices to connect to"
                " synapses that have been created with a connection pattern. You can"
                " avoid this situation by either using a connection pattern or synaptic"
                " indices in both connect calls.",
                name_suffix="cannot_check_synapse_indices",
            )
        n = np.atleast_1d(n)
        p = np.atleast_1d(p)

        if not len(p) == 1 or p != 1:
            use_connections = np.random.rand(len(sources)) < p
            sources = sources[use_connections]
            targets = targets[use_connections]
            n = n[use_connections]
        sources = sources.repeat(n)
        targets = targets.repeat(n)

        variables.add_array(
            "sources", len(sources), dtype=np.int32, values=sources, read_only=True
        )
        variables.add_array(
            "targets", len(targets), dtype=np.int32, values=targets, read_only=True
        )
        # These definitions are important to get the types right in C++
        variables.add_auxiliary_variable("_real_sources", dtype=np.int32)
        variables.add_auxiliary_variable("_real_targets", dtype=np.int32)
        abstract_code = ""
        if "_offset" in self.source.variables:
            variables.add_reference("_source_offset", self.source, "_offset")
            abstract_code += "_real_sources = sources + _source_offset\n"
        else:
            abstract_code += "_real_sources = sources\n"
        if "_offset" in self.target.variables:
            variables.add_reference("_target_offset", self.target, "_offset")
            abstract_code += "_real_targets = targets + _target_offset\n"
        else:
            abstract_code += "_real_targets = targets"
        logger.debug(
            f"Creating synapses from group '{self.source.name}' to group "
            f"'{self.target.name}', using pre-defined arrays)"
        )

        codeobj = create_runner_codeobj(
            self,
            abstract_code,
            "synapses_create_array",
            additional_variables=variables,
            template_kwds=template_kwds,
            needed_variables=needed_variables,
            check_units=False,
            run_namespace={},
        )
        codeobj()

    def _expression_index_dependence(self, expr, namespace, additional_indices=None):
        """
        Returns the set of synaptic indices that expr depends on
        """
        nr = NodeRenderer()
        expr = nr.render_expr(expr)
        deps = set()
        if additional_indices is None:
            additional_indices = {}
        identifiers = get_identifiers_recursively([expr], self.variables)
        variables = self.resolve_all(
            {name for name in identifiers if name not in additional_indices}, namespace
        )
        if any(getattr(var, "auto_vectorise", False) for var in variables.values()):
            identifiers.add("_vectorisation_idx")

        for varname in identifiers:
            # Special handling of i and j -- they do not actually use pre-/
            # postsynaptic indices (except for subgroups), they *are* the
            # pre-/postsynaptic indices
            if varname == "i":
                deps.add("_presynaptic_idx")
            elif varname == "j":
                deps.add("_iterator_idx")
            elif varname in additional_indices:
                deps.add(additional_indices[varname])
            else:
                deps.add(self.variables.indices[varname])
        if "0" in deps:
            deps.remove("0")
        return deps

    def _add_synapses_generator(
        self,
        gen,
        n,
        skip_if_invalid=False,
        over_presynaptic=True,
        namespace=None,
        level=0,
    ):
        # Get the local namespace
        if namespace is None:
            namespace = get_local_namespace(level=level + 1)

        parsed = parse_synapse_generator(gen)
        self._check_parsed_synapses_generator(parsed, namespace)

        # Referring to N_incoming/N_outgoing in the connect statement is
        # ill-defined (see github issue #1227)
        identifiers = get_identifiers_recursively([gen], self.variables)
        for var in ["N_incoming", "N_outgoing"]:
            if var in identifiers:
                raise ValueError(f"The connect statement cannot refer to '{var}'.")

        template_kwds, needed_variables = self._get_multisynaptic_indices()
        template_kwds.update(parsed)
        template_kwds["skip_if_invalid"] = skip_if_invalid
        # To support both i='...' and j='...' syntax, we provide additional keywords
        # to the template
        outer_index = "i" if over_presynaptic else "j"
        outer_index_size = "N_pre" if over_presynaptic else "N_post"
        outer_index_array = "_pre_idx" if over_presynaptic else "_post_idx"
        outer_index_offset = "_source_offset" if over_presynaptic else "_target_offset"
        result_index = "j" if over_presynaptic else "i"
        result_index_size = "N_post" if over_presynaptic else "N_pre"
        target_idx = "_postsynaptic_idx" if over_presynaptic else "_presynaptic_idx"
        result_index_array = "_post_idx" if over_presynaptic else "_pre_idx"
        result_index_offset = "_target_offset" if over_presynaptic else "_source_offset"
        result_index_name = "postsynaptic" if over_presynaptic else "presynaptic"
        template_kwds.update(
            {
                "outer_index": outer_index,
                "outer_index_size": outer_index_size,
                "outer_index_array": outer_index_array,
                "outer_index_offset": outer_index_offset,
                "result_index": result_index,
                "result_index_size": result_index_size,
                "result_index_name": result_index_name,
                "result_index_array": result_index_array,
                "result_index_offset": result_index_offset,
            }
        )
        abstract_code = {
            "setup_iterator": "",
            "generator_expr": "",
            "create_cond": "",
            "update": "",
        }

        additional_indices = {parsed["inner_variable"]: "_iterator_idx"}

        setupiter = ""
        for k, v in parsed["iterator_kwds"].items():
            if v is not None and k != "sample_size":
                deps = self._expression_index_dependence(
                    v, namespace=namespace, additional_indices=additional_indices
                )
                if f"_{result_index_name}_idx" in deps or "_iterator_idx" in deps:
                    raise ValueError(
                        f'Expression "{v}" depends on {result_index_name} '
                        "index or iterator"
                    )
                setupiter += f"_iter_{k} = {v}\n"

        # rand() in the if condition depends on _vectorisation_idx, but not if
        # its in the range expression (handled above)
        additional_indices["_vectorisation_idx"] = "_iterator_idx"

        result_index_condition = False
        result_index_used = False
        if parsed["if_expression"] is not None:
            deps = self._expression_index_dependence(
                parsed["if_expression"],
                namespace=namespace,
                additional_indices=additional_indices,
            )
            if target_idx in deps:
                result_index_condition = True
                result_index_used = True
            elif "_iterator_idx" in deps:
                result_index_condition = True
        template_kwds["result_index_condition"] = result_index_condition
        template_kwds["result_index_used"] = result_index_used

        abstract_code["setup_iterator"] += setupiter
        abstract_code[
            "generator_expr"
        ] += f"{outer_index_array} = _raw{outer_index_array} \n"
        abstract_code["generator_expr"] += f'_{result_index} = {parsed["element"]}\n'

        if result_index_condition:
            abstract_code[
                "create_cond"
            ] += f"{result_index_array} = _raw{result_index_array} \n"
        if parsed["if_expression"] is not None:
            abstract_code["create_cond"] += "_cond = " + parsed["if_expression"] + "\n"
            abstract_code[
                "update"
            ] += f"{result_index_array} = _raw{result_index_array} \n"
        abstract_code["update"] += "_n = " + str(n) + "\n"

        # This overwrites 'i' and 'j' in the synapses' variables dictionary
        # This is necessary because in the context of synapse creation, i
        # and j do not correspond to the sources/targets of the existing
        # synapses but to all the possible sources/targets
        variables = Variables(None)
        # Will be set in the template
        variables.add_auxiliary_variable("_i", dtype=np.int32)
        variables.add_auxiliary_variable("_j", dtype=np.int32)
        variables.add_auxiliary_variable("_iter_low", dtype=np.int32)
        variables.add_auxiliary_variable("_iter_high", dtype=np.int32)
        variables.add_auxiliary_variable("_iter_step", dtype=np.int32)
        variables.add_auxiliary_variable("_iter_p")
        variables.add_auxiliary_variable("_iter_size", dtype=np.int32)
        variables.add_auxiliary_variable(parsed["inner_variable"], dtype=np.int32)
        # Make sure that variables have the correct type in the code
        variables.add_auxiliary_variable("_pre_idx", dtype=np.int32)
        variables.add_auxiliary_variable("_post_idx", dtype=np.int32)
        if parsed["if_expression"] is not None:
            variables.add_auxiliary_variable("_cond", dtype=bool)
        variables.add_auxiliary_variable("_n", dtype=np.int32)

        if "_offset" in self.source.variables:
            variables.add_reference("_source_offset", self.source, "_offset")
        else:
            variables.add_constant("_source_offset", value=0)

        if "_offset" in self.target.variables:
            variables.add_reference("_target_offset", self.target, "_offset")
        else:
            variables.add_constant("_target_offset", value=0)

        variables.add_auxiliary_variable("_raw_pre_idx", dtype=np.int32)
        variables.add_auxiliary_variable("_raw_post_idx", dtype=np.int32)

        variable_indices = defaultdict(lambda: "_idx")
        for varname in self.variables:
            if self.variables.indices[varname] == "_presynaptic_idx":
                variable_indices[varname] = "_raw_pre_idx"
            elif self.variables.indices[varname] == "_postsynaptic_idx":
                variable_indices[varname] = "_raw_post_idx"

        if self.variables["i"] is self.variables["_synaptic_pre"]:
            variables.add_subexpression("i", "_i", dtype=self.variables["i"].dtype)
        if self.variables["j"] is self.variables["_synaptic_post"]:
            variables.add_subexpression("j", "_j", dtype=self.variables["j"].dtype)
        logger.debug(
            f"Creating synapses from group '{self.source.name}' to group "
            f"'{self.target.name}', using generator "
            f"'{parsed['original_expression']}'"
        )

        codeobj = create_runner_codeobj(
            self,
            abstract_code,
            "synapses_create_generator",
            variable_indices=variable_indices,
            additional_variables=variables,
            template_kwds=template_kwds,
            needed_variables=needed_variables,
            check_units=False,
            run_namespace=namespace,
        )
        codeobj()

    def _check_parsed_synapses_generator(self, parsed, namespace):
        """
        Type-check the parsed synapses generator. This function will raise a
        TypeError if any of the arguments to the iterator function are of an
        invalid type.
        """
        if parsed["iterator_func"] == "range":
            # We expect all arguments of the range function to be integers
            for argname, arg in parsed["iterator_kwds"].items():
                identifiers = get_identifiers(arg)
                variables = self.resolve_all(
                    identifiers, run_namespace=namespace, user_identifiers=identifiers
                )
                annotated = brian_ast(arg, variables)
                if annotated.dtype != "integer":
                    raise TypeError(
                        f"The '{argname}' argument of the range function was "
                        f"'{arg}', but it needs to be an integer."
                    )

    def _finalize_generator_expression(
        self,
        generator_expression,
        iteration_index,
        p,
        target_index_name,
        iteration_index_name,
    ):
        if iteration_index is not None:
            raise TypeError(
                f"Generator syntax for {target_index_name} cannot be combined with "
                f"{iteration_index_name} argument"
            )
        if isinstance(p, str) or p != 1:
            raise ValueError("Generator syntax cannot be combined with p argument")
        if not re.search(r"\bfor\b", generator_expression):
            if_split = generator_expression.split(" if ")
            if len(if_split) == 1:
                generator_expression = f"{generator_expression} for _ in range(1)"
            elif len(if_split) == 2:
                generator_expression = (
                    f"{if_split[0]} for _ in range(1) if {if_split[1]}"
                )
            else:
                raise SyntaxError(
                    f"Error parsing expression '{generator_expression}'. "
                    "Expression must have generator "
                    "syntax, for example 'k for k in "
                    f"range({iteration_index_name}-10, {iteration_index_name}+10)'"
                )
        return generator_expression



"""
Example using GSL ODE solvers with a variable time step and comparing it to the
Brian solver.

For highly accurate simulations, i.e. simulations with a very low desired error,
the GSL simulation with a variable time step can be faster because it uses a
low time step only when it is necessary. In biologically detailed models (e.g.
of the Hodgkin-Huxley type), the relevant time constants are very short around
an action potential, but much longer when the neuron is near its resting
potential. The following example uses a very simple neuron model (leaky
integrate-and-fire), but simulates a change in relevant time constants by
changing the actual time constant every 10ms, independently for each of 100
neurons. To accurately simulate this model with a fixed time step, the time step
has to be very small, wasting many unnecessary steps for all the neurons where
the time constant is long.

Note that using the GSL ODE solver is much slower, if both methods use a
comparable number of steps, i.e. if the desired accuracy is low enough so that
a single step per "Brian time step" is enough.
"""
from brian2 import *
import time

# Run settings
start_dt = .1 * ms
method = 'rk2'
error = 1.e-6  # requested accuracy


def runner(method, dt, options=None):
    seed(0)
    I = 5
    group = NeuronGroup(100, '''dv/dt = (-v + I)/tau : 1
                                tau : second''',
                        method=method,
                        method_options=options,
                        dt=dt)
    group.run_regularly('''v = rand()
                           tau = 0.1*ms + rand()*9.9*ms''', dt=10*ms)

    rec_vars = ['v', 'tau']
    if 'gsl' in method:
        rec_vars += ['_step_count']
    net = Network(group)
    net.run(0 * ms)
    mon = StateMonitor(group, rec_vars, record=True, dt=start_dt)
    net.add(mon)
    start = time.time()
    net.run(1 * second)
    mon.add_attribute('run_time')
    mon.run_time = time.time() - start
    return mon


lin = runner('linear', start_dt)
method_options = {'save_step_count': True,
                  'absolute_error': error,
                  'max_steps': 10000}
gsl = runner('gsl_%s' % method, start_dt, options=method_options)

print("Running with GSL integrator and variable time step:")
print('Run time: %.3fs' % gsl.run_time)

# check gsl error
assert np.max(np.abs(
    lin.v - gsl.v)) < error, "Maximum error gsl integration too large: %f" % np.max(
    np.abs(lin.v - gsl.v))
print("average step count: %.1f" % np.mean(gsl._step_count))
print("average absolute error: %g" % np.mean(np.abs(gsl.v - lin.v)))

print("\nRunning with exact integration and fixed time step:")
dt = start_dt
count = 0
dts = []
avg_errors = []
max_errors = []
runtimes = []
while True:
    print('Using dt: %s' % str(dt))
    brian = runner(method, dt)
    print('\tRun time: %.3fs' % brian.run_time)
    avg_errors.append(np.mean(np.abs(brian.v - lin.v)))
    max_errors.append(np.max(np.abs(brian.v - lin.v)))
    dts.append(dt)
    runtimes.append(brian.run_time)
    if np.max(np.abs(brian.v - lin.v)) > error:
        print('\tError too high (%g), decreasing dt' % np.max(
            np.abs(brian.v - lin.v)))
        dt *= .5
        count += 1
    else:
        break
print("Desired error level achieved:")
print("average step count: %.2fs" % (start_dt / dt))
print("average absolute error: %g" % np.mean(np.abs(brian.v - lin.v)))

print('Run time: %.3fs' % brian.run_time)
if brian.run_time > gsl.run_time:
    print("This is %.1f times slower than the simulation with GSL's variable "
          "time step method." % (brian.run_time / gsl.run_time))
else:
    print("This is %.1f times faster than the simulation with GSL's variable "
          "time step method." % (gsl.run_time / brian.run_time))

fig, (ax1, ax2) = plt.subplots(1, 2)
ax2.axvline(1e-6, color='gray')
for label, gsl_error, std_errors, ax in [('average absolute error', np.mean(np.abs(gsl.v - lin.v)), avg_errors, ax1),
                                         ('maximum absolute error', np.max(np.abs(gsl.v - lin.v)), max_errors, ax2)]:
    ax.set(xscale='log', yscale='log')
    ax.plot([], [], 'o', color='C0', label='fixed time step')  # for the legend entry
    for (error, runtime, dt) in zip(std_errors, runtimes, dts):
        ax.plot(error, runtime, 'o', color='C0')
        ax.annotate('%s' % str(dt), xy=(error, runtime), xytext=(2.5, 5),
                    textcoords='offset points', color='C0')
    ax.plot(gsl_error, gsl.run_time, 'o', color='C1', label='variable time step (GSL)')
    ax.set(xlabel=label, xlim=(10**-10, 10**1))
ax1.set_ylabel('runtime (s)')
ax2.legend(loc='lower left')

plt.show()



#!/usr/bin/env python
'''
Spike-timing dependent plasticity
Adapted from Song, Miller and Abbott (2000) and Song and Abbott (2001)

This example is modified from ``synapses_STDP.py`` and writes a standalone C++ project in the directory
``STDP_standalone``.
'''
import sys, time, os
from brian2 import *

standalone = int(sys.argv[-2])
n_threads  = int(sys.argv[-1])
path       = 'data_stdp_%d' %n_threads

if standalone == 1:
    set_device('cpp_standalone')
    brian_prefs.codegen.cpp_standalone.openmp_threads = n_threads

start   = time.time()
N       = 1000
taum    = 10 * ms
taupre  = 20 * ms
taupost = taupre
Ee      = 0 * mV
vt      = -54 * mV
vr      = -60 * mV
El      = -74 * mV
taue    = 5 * ms
F       =  30 * Hz
gmax    = .01
dApre   = .01
dApost  = -dApre * taupre / taupost * 1.05
dApost *= gmax
dApre  *= gmax

eqs_neurons = '''
dv/dt=(ge*(Ee-vr)+El-v)/taum : volt   # the synaptic current is linearized
dge/dt=-ge/taue : 1
'''

poisson_input   = PoissonGroup(N, rates=F)
neurons = NeuronGroup(500, eqs_neurons, threshold='v>vt', reset='v=vr')
S = Synapses(poisson_input, neurons,
             '''w:1
                dApre/dt=-Apre/taupre : 1 (event-driven)    
                dApost/dt=-Apost/taupost : 1 (event-driven)''',
             on_pre='''ge+=w
                    Apre+=dApre
                    w=clip(w+Apost,0,gmax)''',
             on_post='''Apost+=dApost
                     w=clip(w+Apre,0,gmax)''')
S.connect()
S.w          = 'rand()*gmax'
state_mon    = StateMonitor(S, 'w', record=[0])
spike_mon_1  = SpikeMonitor(poisson_input)
spike_mon_2  = SpikeMonitor(neurons)
start_time   = time.time()

net = Network(poisson_input, neurons, S, state_mon, spike_mon_1, spike_mon_2, name='stdp_net')

if standalone == 1:
    device.insert_code('main', 'std::clock_t start = std::clock();')

net.run(5 * second, report='text')

if standalone == 1:
    device.insert_code('main', '''
        std::ofstream myfile ("speed.txt");
        if (myfile.is_open())
        {
            double value = (double) (std::clock() - start)/(%d * CLOCKS_PER_SEC); 
            myfile << value << std::endl;
            myfile.close();
        }
        ''' %(max(1, n_threads)))

try:
    os.removedirs(path)
except Exception:
    pass

if standalone == 1:
    device.build(project_dir=path, compile_project=True, run_project=True, debug=False)




"""
This is only a bridge for using Brian 1 hears with Brian 2.

.. deprecated:: 2.2.2.2
    Use the `brian2hears <https://brian2hears.readthedocs.io/>`_ package instead.

NOTES:

* Slicing sounds with Brian 2 units doesn't work, you need to either use Brian 1 units or replace calls to
  ``sound[:20*ms]`` with ``sound.slice(None, 20*ms)``, etc.

TODO: handle properties (e.g. sound.duration)

Not working examples:

* time_varying_filter1 (care with units)
"""

try:
    import brian as b1
    import brian.hears as b1h
except ImportError:
    raise ImportError(
        "brian2.hears is deprecated and will be removed in a future release, please use"
        " the brian2hears package available at https://brian2hears.readthedocs.io/. If"
        " you really want to keep using it, note: brian2.hears is a bridge between"
        " Brian 2 and the version of Brian Hears from Brian 1, you need to have Brian 1"
        " installed to use it."
    )

from inspect import isclass, ismethod

from numpy import asarray, ndarray

from brian2.core.clocks import Clock
from brian2.core.operations import network_operation
from brian2.groups.neurongroup import NeuronGroup
from brian2.units import second
from brian2.units.fundamentalunits import Quantity
from brian2.utils.logger import get_logger

logger = get_logger(__name__)

logger.warn(
    "brian2.hears is deprecated and will be removed in a future release, please use the"
    " brian2hears package available at https://brian2hears.readthedocs.io/. If you"
    " really want to keep using it, note that it is a bridge between Brian 2 and Brian"
    " Hears from Brian 1. This is not guaranteed to work in all cases that brian.hears"
    " works. See the limitations in the online documentation."
)


def convert_unit_b1_to_b2(val):
    return Quantity.with_dimensions(float(val), val.dim._dims)


def convert_unit_b2_to_b1(val):
    return b1.Quantity.with_dimensions(float(val), val.dim._dims)


def modify_arg(arg):
    """
    Modify arguments to make them compatible with Brian 1.

    - Arrays of units are replaced with straight arrays
    - Single values are replaced with Brian 1 equivalents
    - Slices are handled so we can use e.g. sound[:20*ms]

    The second part was necessary because some functions/classes test if an object is an array or not to see if it
    is a sequence, but because brian2.Quantity derives from ndarray this was causing problems.
    """
    if isinstance(arg, Quantity):
        if len(arg.shape) == 0:
            arg = b1.Quantity.with_dimensions(float(arg), arg.dim._dims)
        else:
            arg = asarray(arg)
    elif isinstance(arg, slice):
        arg = slice(modify_arg(arg.start), modify_arg(arg.stop), modify_arg(arg.step))
    return arg


def wrap_units(f):
    """
    Wrap a function to convert units into a form that Brian 1 can handle. Also, check the output argument, if it is
    a ``b1h.Sound`` wrap it.
    """

    def new_f(*args, **kwds):
        newargs = []
        newkwds = {}
        for arg in args:
            newargs.append(modify_arg(arg))
        for k, v in kwds.items():
            newkwds[k] = modify_arg(v)
        rv = f(*newargs, **newkwds)
        if rv.__class__ == b1h.Sound:
            rv.__class__ = BridgeSound
        elif isinstance(rv, b1.Quantity):
            rv = Quantity.with_dimensions(float(rv), rv.dim._dims)
        return rv

    return new_f


def wrap_units_property(p):
    fget = p.fget
    fset = p.fset
    fdel = p.fdel
    if fget is not None:
        fget = wrap_units(fget)
    if fset is not None:
        fset = wrap_units(fset)
    if fdel is not None:
        fdel = wrap_units(fdel)
    new_p = property(fget, fset, fdel)
    return new_p


def wrap_units_class(_C):
    """
    Wrap a class to convert units into a form that Brian 1 can handle in all methods
    """

    class new_class(_C):
        for _k in _C.__dict__:
            _v = getattr(_C, _k)
            if hasattr(ndarray, _k) and getattr(ndarray, _k) is _v:
                continue
            if ismethod(_v):
                _v = wrap_units(_v)
                exec(f"{_k} = _v")
            elif isinstance(_v, property):
                _v = wrap_units_property(_v)
                exec(f"{_k} = _v")
        del _k
        del _v

    return new_class


WrappedSound = wrap_units_class(b1h.Sound)


class BridgeSound(WrappedSound):
    """
    We add a new method slice because slicing with units can't work with Brian 2 units.
    """

    def slice(self, *args):
        return self.__getitem__(slice(*args))


Sound = BridgeSound


class FilterbankGroup(NeuronGroup):
    def __init__(self, filterbank, targetvar, *args, **kwds):
        self.targetvar = targetvar
        self.filterbank = filterbank
        self.buffer = None
        filterbank.buffer_init()

        # Sanitize the clock - does it have the right dt value?
        if "clock" in kwds:
            if int(1 / kwds["clock"].dt) != int(filterbank.samplerate):
                raise ValueError("Clock should have 1/dt=samplerate")
            kwds["clock"] = Clock(dt=float(kwds["clock"].dt) * second)
        else:
            kwds["clock"] = Clock(dt=1 * second / float(filterbank.samplerate))

        buffersize = kwds.pop("buffersize", 32)
        if not isinstance(buffersize, int):
            buffersize = int(buffersize * self.samplerate)
        self.buffersize = buffersize
        self.buffer_pointer = buffersize
        self.buffer_start = -buffersize

        NeuronGroup.__init__(self, filterbank.nchannels, *args, **kwds)

        @network_operation(clock=self.clock, when="start")
        def apply_filterbank_output():
            if self.buffer_pointer >= self.buffersize:
                self.buffer_pointer = 0
                self.buffer_start += self.buffersize
                self.buffer = self.filterbank.buffer_fetch(
                    self.buffer_start, self.buffer_start + self.buffersize
                )
            setattr(self, targetvar, self.buffer[self.buffer_pointer, :])
            self.buffer_pointer += 1

        self.contained_objects.append(apply_filterbank_output)

    def reinit(self):
        NeuronGroup.reinit(self)
        self.filterbank.buffer_init()
        self.buffer_pointer = self.buffersize
        self.buffer_start = -self.buffersize


handled_explicitly = {"Sound", "FilterbankGroup"}

__all__ = [k for k in b1h.__dict__ if not k.startswith("_")]
for k in __all__:
    if k in handled_explicitly:
        continue
    curobj = getattr(b1h, k)
    if callable(curobj):
        if isclass(curobj):
            curobj = wrap_units_class(curobj)
        else:
            curobj = wrap_units(curobj)
    exec(f"{k} = curobj")

__all__.extend(
    [
        "convert_unit_b1_to_b2",
        "convert_unit_b2_to_b1",
    ]
)



"""
Utility functions for handling the units in `Equations`.
"""

import re

from brian2.core.variables import Variable
from brian2.parsing.expressions import parse_expression_dimensions
from brian2.parsing.statements import parse_statement
from brian2.units.fundamentalunits import (
    fail_for_dimension_mismatch,
    get_dimensions,
    get_unit,
)

__all__ = ["check_dimensions", "check_units_statements"]


def check_dimensions(expression, dimensions, variables):
    """
    Compares the physical dimensions of an expression to expected dimensions in
    a given namespace.

    Parameters
    ----------
    expression : str
        The expression to evaluate.
    dimensions : `Dimension`
        The expected physical dimensions for the `expression`.
    variables : dict
        Dictionary of all variables (including external constants) used in
        the `expression`.

    Raises
    ------
    KeyError
        In case on of the identifiers cannot be resolved.
    DimensionMismatchError
        If an unit mismatch occurs during the evaluation.
    """
    expr_dims = parse_expression_dimensions(expression, variables)
    expected = repr(get_unit(dimensions))
    err_msg = (
        f"Expression '{expression.strip()}' does not have the expected unit {expected}"
    )
    fail_for_dimension_mismatch(expr_dims, dimensions, err_msg)


def check_units_statements(code, variables):
    """
    Check the units for a series of statements. Setting a model variable has to
    use the correct unit. For newly introduced temporary variables, the unit
    is determined and used to check the following statements to ensure
    consistency.

    Parameters
    ----------
    code : str
        The statements as a (multi-line) string
    variables : dict of `Variable` objects
        The information about all variables used in `code` (including
        `Constant` objects for external variables)

    Raises
    ------
    KeyError
        In case on of the identifiers cannot be resolved.
    DimensionMismatchError
        If an unit mismatch occurs during the evaluation.
    """
    variables = dict(variables)
    # Avoid a circular import
    from brian2.codegen.translation import analyse_identifiers

    newly_defined, _, unknown = analyse_identifiers(code, variables)

    if len(unknown):
        raise AssertionError(
            "Encountered unknown identifiers, this should not "
            f"happen at this stage. Unknown identifiers: {unknown}"
        )

    code = re.split(r"[;\n]", code)
    for line in code:
        line = line.strip()
        if not len(line):
            continue  # skip empty lines

        varname, op, expr, comment = parse_statement(line)
        if op in ("+=", "-=", "*=", "/=", "%="):
            # Replace statements such as "w *=2" by "w = w * 2"
            expr = f"{varname} {op[0]} {expr}"
        elif op == "=":
            pass
        else:
            raise AssertionError(f'Unknown operator "{op}"')

        expr_unit = parse_expression_dimensions(expr, variables)

        if varname in variables:
            expected_unit = variables[varname].dim
            fail_for_dimension_mismatch(
                expr_unit,
                expected_unit,
                "The right-hand-side of code "
                'statement "%s" does not have the '
                "expected unit {expected}" % line,
                expected=expected_unit,
            )
        elif varname in newly_defined:
            # note the unit for later
            variables[varname] = Variable(
                name=varname, dimensions=get_dimensions(expr_unit), scalar=False
            )
        else:
            raise AssertionError(
                f"Variable '{varname}' is neither in the variables "
                "dictionary nor in the list of undefined "
                "variables."
            )
